# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests


import cartopy.crs as ccrs

import iris.cube
import iris.coords
import iris.tests.stock

from iris.coord_systems import (
    GeogCS,
    LambertConformal,
    RotatedGeogCS,
    Stereographic,
    TransverseMercator,
)


def osgb():
    return TransverseMercator(
        latitude_of_projection_origin=49,
        longitude_of_central_meridian=-2,
        false_easting=-400,
        false_northing=100,
        scale_factor_at_central_meridian=0.9996012717,
        ellipsoid=GeogCS(6377563.396, 6356256.909),
    )


def stereo():
    return Stereographic(
        central_lat=-90,
        central_lon=-45,
        false_easting=100,
        false_northing=200,
        ellipsoid=GeogCS(6377563.396, 6356256.909),
    )


class TestCoordSystemLookup(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.lat_lon_cube()

    def test_hit_name(self):
        self.assertIsInstance(self.cube.coord_system("GeogCS"), GeogCS)

    def test_hit_type(self):
        self.assertIsInstance(self.cube.coord_system(GeogCS), GeogCS)

    def test_miss(self):
        self.assertIsNone(self.cube.coord_system(RotatedGeogCS))

    def test_empty(self):
        self.assertIsInstance(self.cube.coord_system(GeogCS), GeogCS)
        self.assertIsNotNone(self.cube.coord_system(None))
        self.assertIsInstance(self.cube.coord_system(None), GeogCS)
        self.assertIsNotNone(self.cube.coord_system())
        self.assertIsInstance(self.cube.coord_system(), GeogCS)

        for coord in self.cube.coords():
            coord.coord_system = None

        self.assertIsNone(self.cube.coord_system(GeogCS))
        self.assertIsNone(self.cube.coord_system(None))
        self.assertIsNone(self.cube.coord_system())


class TestCoordSystemSame(tests.IrisTest):
    def setUp(self):
        self.cs1 = iris.coord_systems.GeogCS(6371229)
        self.cs2 = iris.coord_systems.GeogCS(6371229)
        self.cs3 = iris.coord_systems.RotatedGeogCS(
            30, 30, ellipsoid=GeogCS(6371229)
        )

    def test_simple(self):
        a = self.cs1
        b = self.cs2
        self.assertEqual(a, b)

    def test_different_class(self):
        a = self.cs1
        b = self.cs3
        self.assertNotEquals(a, b)

    def test_different_public_attributes(self):
        a = self.cs1
        b = self.cs2
        a.foo = "a"

        # check that that attribute was added (just in case)
        self.assertEqual(a.foo, "a")

        # a and b should not be the same
        self.assertNotEquals(a, b)

        # a and b should be the same
        b.foo = "a"
        self.assertEqual(a, b)

        b.foo = "b"
        # a and b should not be the same
        self.assertNotEquals(a, b)


class Test_CoordSystem_xml_element(tests.IrisTest):
    def test_rotated(self):
        cs = RotatedGeogCS(30, 40, ellipsoid=GeogCS(6371229))
        self.assertXMLElement(
            cs, ("coord_systems", "CoordSystem_xml_element.xml")
        )


class Test_GeogCS_construction(tests.IrisTest):
    # Test Ellipsoid constructor
    # Don't care about testing the units, it has no logic specific to this class.

    def test_sphere_param(self):
        cs = GeogCS(6543210)
        self.assertXMLElement(cs, ("coord_systems", "GeogCS_init_sphere.xml"))

    def test_no_major(self):
        cs = GeogCS(
            semi_minor_axis=6500000, inverse_flattening=151.42814163388104
        )
        self.assertXMLElement(
            cs, ("coord_systems", "GeogCS_init_no_major.xml")
        )

    def test_no_minor(self):
        cs = GeogCS(
            semi_major_axis=6543210, inverse_flattening=151.42814163388104
        )
        self.assertXMLElement(
            cs, ("coord_systems", "GeogCS_init_no_minor.xml")
        )

    def test_no_invf(self):
        cs = GeogCS(semi_major_axis=6543210, semi_minor_axis=6500000)
        self.assertXMLElement(cs, ("coord_systems", "GeogCS_init_no_invf.xml"))

    def test_invalid_ellipsoid_params(self):
        # no params
        with self.assertRaises(ValueError):
            GeogCS()

        # over specified
        with self.assertRaises(ValueError):
            GeogCS(6543210, 6500000, 151.42814163388104)

        # under specified
        with self.assertRaises(ValueError):
            GeogCS(None, 6500000, None)
        with self.assertRaises(ValueError):
            GeogCS(None, None, 151.42814163388104)


class Test_GeogCS_repr(tests.IrisTest):
    def test_repr(self):
        cs = GeogCS(6543210, 6500000)
        expected = (
            "GeogCS(semi_major_axis=6543210.0, semi_minor_axis=6500000.0)"
        )
        self.assertEqual(expected, repr(cs))


class Test_GeogCS_str(tests.IrisTest):
    def test_str(self):
        cs = GeogCS(6543210, 6500000)
        expected = (
            "GeogCS(semi_major_axis=6543210.0, semi_minor_axis=6500000.0)"
        )
        self.assertEqual(expected, str(cs))


class Test_GeogCS_as_cartopy_globe(tests.IrisTest):
    def test_as_cartopy_globe(self):
        cs = GeogCS(6543210, 6500000)
        # Can't check equality directly, so use the proj4 params instead.
        res = cs.as_cartopy_globe().to_proj4_params()
        expected = {"a": 6543210, "b": 6500000}
        self.assertEqual(res, expected)


class Test_GeogCS_as_cartopy_crs(tests.IrisTest):
    def test_as_cartopy_crs(self):
        cs = GeogCS(6543210, 6500000)
        res = cs.as_cartopy_crs()
        globe = ccrs.Globe(
            semimajor_axis=6543210.0, semiminor_axis=6500000.0, ellipse=None
        )
        expected = ccrs.Geodetic(globe)
        self.assertEqual(res, expected)


class Test_RotatedGeogCS_construction(tests.IrisTest):
    def test_init(self):
        rcs = RotatedGeogCS(
            30, 40, north_pole_grid_longitude=50, ellipsoid=GeogCS(6371229)
        )
        self.assertXMLElement(rcs, ("coord_systems", "RotatedGeogCS_init.xml"))

        rcs = RotatedGeogCS(30, 40, north_pole_grid_longitude=50)
        self.assertXMLElement(
            rcs, ("coord_systems", "RotatedGeogCS_init_a.xml")
        )

        rcs = RotatedGeogCS(30, 40)
        self.assertXMLElement(
            rcs, ("coord_systems", "RotatedGeogCS_init_b.xml")
        )


class Test_RotatedGeogCS_repr(tests.IrisTest):
    def test_repr(self):
        rcs = RotatedGeogCS(
            30, 40, north_pole_grid_longitude=50, ellipsoid=GeogCS(6371229)
        )
        expected = (
            "RotatedGeogCS(30.0, 40.0, "
            "north_pole_grid_longitude=50.0, ellipsoid=GeogCS(6371229.0))"
        )
        self.assertEqual(expected, repr(rcs))

        rcs = RotatedGeogCS(30, 40, north_pole_grid_longitude=50)
        expected = "RotatedGeogCS(30.0, 40.0, north_pole_grid_longitude=50.0)"
        self.assertEqual(expected, repr(rcs))

        rcs = RotatedGeogCS(30, 40)
        expected = "RotatedGeogCS(30.0, 40.0)"
        self.assertEqual(expected, repr(rcs))


class Test_RotatedGeogCS_str(tests.IrisTest):
    def test_str(self):
        rcs = RotatedGeogCS(
            30, 40, north_pole_grid_longitude=50, ellipsoid=GeogCS(6371229)
        )
        expected = (
            "RotatedGeogCS(30.0, 40.0, "
            "north_pole_grid_longitude=50.0, ellipsoid=GeogCS(6371229.0))"
        )
        self.assertEqual(expected, str(rcs))

        rcs = RotatedGeogCS(30, 40, north_pole_grid_longitude=50)
        expected = "RotatedGeogCS(30.0, 40.0, north_pole_grid_longitude=50.0)"
        self.assertEqual(expected, str(rcs))

        rcs = RotatedGeogCS(30, 40)
        expected = "RotatedGeogCS(30.0, 40.0)"
        self.assertEqual(expected, str(rcs))


class Test_TransverseMercator_construction(tests.IrisTest):
    def test_osgb(self):
        tm = osgb()
        self.assertXMLElement(
            tm, ("coord_systems", "TransverseMercator_osgb.xml")
        )


class Test_TransverseMercator_repr(tests.IrisTest):
    def test_osgb(self):
        tm = osgb()
        expected = (
            "TransverseMercator(latitude_of_projection_origin=49.0, longitude_of_central_meridian=-2.0, "
            "false_easting=-400.0, false_northing=100.0, scale_factor_at_central_meridian=0.9996012717, "
            "ellipsoid=GeogCS(semi_major_axis=6377563.396, semi_minor_axis=6356256.909))"
        )
        self.assertEqual(expected, repr(tm))


class Test_TransverseMercator_as_cartopy_crs(tests.IrisTest):
    def test_as_cartopy_crs(self):
        latitude_of_projection_origin = 49.0
        longitude_of_central_meridian = -2.0
        false_easting = -40000.0
        false_northing = 10000.0
        scale_factor_at_central_meridian = 0.9996012717
        ellipsoid = GeogCS(
            semi_major_axis=6377563.396, semi_minor_axis=6356256.909
        )

        tmerc_cs = TransverseMercator(
            latitude_of_projection_origin,
            longitude_of_central_meridian,
            false_easting,
            false_northing,
            scale_factor_at_central_meridian,
            ellipsoid=ellipsoid,
        )

        expected = ccrs.TransverseMercator(
            central_longitude=longitude_of_central_meridian,
            central_latitude=latitude_of_projection_origin,
            false_easting=false_easting,
            false_northing=false_northing,
            scale_factor=scale_factor_at_central_meridian,
            globe=ccrs.Globe(
                semimajor_axis=6377563.396,
                semiminor_axis=6356256.909,
                ellipse=None,
            ),
        )

        res = tmerc_cs.as_cartopy_crs()
        self.assertEqual(res, expected)


class Test_TransverseMercator_as_cartopy_projection(tests.IrisTest):
    def test_as_cartopy_projection(self):
        latitude_of_projection_origin = 49.0
        longitude_of_central_meridian = -2.0
        false_easting = -40000.0
        false_northing = 10000.0
        scale_factor_at_central_meridian = 0.9996012717
        ellipsoid = GeogCS(
            semi_major_axis=6377563.396, semi_minor_axis=6356256.909
        )

        tmerc_cs = TransverseMercator(
            latitude_of_projection_origin,
            longitude_of_central_meridian,
            false_easting,
            false_northing,
            scale_factor_at_central_meridian,
            ellipsoid=ellipsoid,
        )

        expected = ccrs.TransverseMercator(
            central_longitude=longitude_of_central_meridian,
            central_latitude=latitude_of_projection_origin,
            false_easting=false_easting,
            false_northing=false_northing,
            scale_factor=scale_factor_at_central_meridian,
            globe=ccrs.Globe(
                semimajor_axis=6377563.396,
                semiminor_axis=6356256.909,
                ellipse=None,
            ),
        )

        res = tmerc_cs.as_cartopy_projection()
        self.assertEqual(res, expected)


class Test_Stereographic_construction(tests.IrisTest):
    def test_stereo(self):
        st = stereo()
        self.assertXMLElement(st, ("coord_systems", "Stereographic.xml"))


class Test_Stereographic_repr(tests.IrisTest):
    def test_stereo(self):
        st = stereo()
        expected = (
            "Stereographic(central_lat=-90.0, central_lon=-45.0, "
            "false_easting=100.0, false_northing=200.0, true_scale_lat=None, "
            "ellipsoid=GeogCS(semi_major_axis=6377563.396, semi_minor_axis=6356256.909))"
        )
        self.assertEqual(expected, repr(st))


class Test_Stereographic_as_cartopy_crs(tests.IrisTest):
    def test_as_cartopy_crs(self):
        latitude_of_projection_origin = -90.0
        longitude_of_projection_origin = -45.0
        false_easting = 100.0
        false_northing = 200.0
        ellipsoid = GeogCS(6377563.396, 6356256.909)

        st = Stereographic(
            central_lat=latitude_of_projection_origin,
            central_lon=longitude_of_projection_origin,
            false_easting=false_easting,
            false_northing=false_northing,
            ellipsoid=ellipsoid,
        )
        expected = ccrs.Stereographic(
            central_latitude=latitude_of_projection_origin,
            central_longitude=longitude_of_projection_origin,
            false_easting=false_easting,
            false_northing=false_northing,
            globe=ccrs.Globe(
                semimajor_axis=6377563.396,
                semiminor_axis=6356256.909,
                ellipse=None,
            ),
        )

        res = st.as_cartopy_crs()
        self.assertEqual(res, expected)


class Test_Stereographic_as_cartopy_projection(tests.IrisTest):
    def test_as_cartopy_projection(self):
        latitude_of_projection_origin = -90.0
        longitude_of_projection_origin = -45.0
        false_easting = 100.0
        false_northing = 200.0
        ellipsoid = GeogCS(6377563.396, 6356256.909)

        st = Stereographic(
            central_lat=latitude_of_projection_origin,
            central_lon=longitude_of_projection_origin,
            false_easting=false_easting,
            false_northing=false_northing,
            ellipsoid=ellipsoid,
        )
        expected = ccrs.Stereographic(
            central_latitude=latitude_of_projection_origin,
            central_longitude=longitude_of_projection_origin,
            false_easting=false_easting,
            false_northing=false_northing,
            globe=ccrs.Globe(
                semimajor_axis=6377563.396,
                semiminor_axis=6356256.909,
                ellipse=None,
            ),
        )

        res = st.as_cartopy_projection()
        self.assertEqual(res, expected)


class Test_LambertConformal(tests.GraphicsTest):
    def test_fail_secant_latitudes_none(self):
        emsg = "secant latitudes"
        with self.assertRaisesRegex(ValueError, emsg):
            LambertConformal(secant_latitudes=())

    def test_fail_secant_latitudes_excessive(self):
        emsg = "secant latitudes"
        with self.assertRaisesRegex(ValueError, emsg):
            LambertConformal(secant_latitudes=(1, 2, 3))

    def test_secant_latitudes_single_value(self):
        lat_1 = 40
        lcc = LambertConformal(secant_latitudes=lat_1)
        ccrs = lcc.as_cartopy_crs()
        self.assertEqual(lat_1, ccrs.proj4_params["lat_1"])
        self.assertNotIn("lat_2", ccrs.proj4_params)

    def test_secant_latitudes(self):
        lat_1, lat_2 = 40, 41
        lcc = LambertConformal(secant_latitudes=(lat_1, lat_2))
        ccrs = lcc.as_cartopy_crs()
        self.assertEqual(lat_1, ccrs.proj4_params["lat_1"])
        self.assertEqual(lat_2, ccrs.proj4_params["lat_2"])

    def test_north_cutoff(self):
        lcc = LambertConformal(0, 0, secant_latitudes=(30, 60))
        ccrs = lcc.as_cartopy_crs()
        self.assertEqual(ccrs.cutoff, -30)

    def test_south_cutoff(self):
        lcc = LambertConformal(0, 0, secant_latitudes=(-30, -60))
        ccrs = lcc.as_cartopy_crs()
        self.assertEqual(ccrs.cutoff, 30)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test cube indexing, slicing, and extracting, and also the dot graphs.

"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import collections
import os
import re

import cf_units
import numpy as np
import numpy.ma as ma

import iris
import iris.analysis
import iris.coords
import iris.cube
import iris.fileformats
import iris.fileformats.dot
import iris.tests.pp as pp
import iris.tests.stock


class IrisDotTest(tests.IrisTest):
    def check_dot(self, cube, reference_filename):
        test_string = iris.fileformats.dot.cube_text(cube)
        reference_path = tests.get_result_path(reference_filename)
        if os.path.isfile(reference_path):
            with open(reference_path, "r") as reference_fh:
                reference = "".join(reference_fh.readlines())
            self._assert_str_same(
                reference,
                test_string,
                reference_filename,
                type_comparison_name="DOT files",
            )
        else:
            with open(reference_path, "w") as reference_fh:
                reference_fh.writelines(test_string)


class TestBasicCubeConstruction(tests.IrisTest):
    def setUp(self):
        self.cube = iris.cube.Cube(
            np.arange(12, dtype=np.int32).reshape((3, 4)),
            long_name="test cube",
        )
        self.x = iris.coords.DimCoord(
            np.array([-7.5, 7.5, 22.5, 37.5]), long_name="x"
        )
        self.y = iris.coords.DimCoord(
            np.array([2.5, 7.5, 12.5]), long_name="y"
        )
        self.xy = iris.coords.AuxCoord(
            np.arange(12).reshape((3, 4)) * 3.0, long_name="xy"
        )

    def test_add_dim_coord(self):
        # Lengths must match
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(self.y, 1)
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(self.x, 0)

        # Must specify a dimension
        with self.assertRaises(TypeError):
            self.cube.add_dim_coord(self.y)

        # Add y
        self.cube.add_dim_coord(self.y, 0)
        self.assertEqual(self.cube.coords(), [self.y])
        self.assertEqual(self.cube.dim_coords, (self.y,))
        # Add x
        self.cube.add_dim_coord(self.x, 1)
        self.assertEqual(self.cube.coords(), [self.y, self.x])
        self.assertEqual(self.cube.dim_coords, (self.y, self.x))

        # Cannot add a coord twice
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(self.y, 0)
        # ... even to cube.aux_coords
        with self.assertRaises(ValueError):
            self.cube.add_aux_coord(self.y, 0)

        # Can't add AuxCoord to dim_coords
        y_other = iris.coords.AuxCoord(
            np.array([2.5, 7.5, 12.5]), long_name="y_other"
        )
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(y_other, 0)

    def test_add_scalar_coord(self):
        scalar_dim_coord = iris.coords.DimCoord(
            23, long_name="scalar_dim_coord"
        )
        scalar_aux_coord = iris.coords.AuxCoord(
            23, long_name="scalar_aux_coord"
        )
        # Scalars cannot be in cube.dim_coords
        with self.assertRaises(TypeError):
            self.cube.add_dim_coord(scalar_dim_coord)
        with self.assertRaises(TypeError):
            self.cube.add_dim_coord(scalar_dim_coord, None)
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(scalar_dim_coord, [])
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(scalar_dim_coord, ())

        # Make sure that's still the case for a 0-dimensional cube.
        cube = iris.cube.Cube(666)
        self.assertEqual(cube.ndim, 0)
        with self.assertRaises(TypeError):
            self.cube.add_dim_coord(scalar_dim_coord)
        with self.assertRaises(TypeError):
            self.cube.add_dim_coord(scalar_dim_coord, None)
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(scalar_dim_coord, [])
        with self.assertRaises(ValueError):
            self.cube.add_dim_coord(scalar_dim_coord, ())

        cube = self.cube.copy()
        cube.add_aux_coord(scalar_dim_coord)
        cube.add_aux_coord(scalar_aux_coord)
        self.assertEqual(
            set(cube.aux_coords), {scalar_dim_coord, scalar_aux_coord}
        )

        # Various options for dims
        cube = self.cube.copy()
        cube.add_aux_coord(scalar_dim_coord, [])
        self.assertEqual(cube.aux_coords, (scalar_dim_coord,))

        cube = self.cube.copy()
        cube.add_aux_coord(scalar_dim_coord, ())
        self.assertEqual(cube.aux_coords, (scalar_dim_coord,))

        cube = self.cube.copy()
        cube.add_aux_coord(scalar_dim_coord, None)
        self.assertEqual(cube.aux_coords, (scalar_dim_coord,))

        cube = self.cube.copy()
        cube.add_aux_coord(scalar_dim_coord)
        self.assertEqual(cube.aux_coords, (scalar_dim_coord,))

    def test_add_aux_coord(self):
        y_another = iris.coords.DimCoord(
            np.array([2.5, 7.5, 12.5]), long_name="y_another"
        )

        # DimCoords can live in cube.aux_coords
        self.cube.add_aux_coord(y_another, 0)
        self.assertEqual(self.cube.dim_coords, ())
        self.assertEqual(self.cube.coords(), [y_another])
        self.assertEqual(self.cube.aux_coords, (y_another,))

        # AuxCoords in cube.aux_coords
        self.cube.add_aux_coord(self.xy, [0, 1])
        self.assertEqual(self.cube.dim_coords, ())
        self.assertEqual(self.cube.coords(), [y_another, self.xy])
        self.assertEqual(set(self.cube.aux_coords), {y_another, self.xy})

        # Lengths must match up
        cube = self.cube.copy()
        with self.assertRaises(ValueError):
            cube.add_aux_coord(self.xy, [1, 0])

    def test_remove_coord(self):
        self.cube.add_dim_coord(self.y, 0)
        self.cube.add_dim_coord(self.x, 1)
        self.cube.add_aux_coord(self.xy, (0, 1))
        self.assertEqual(set(self.cube.coords()), {self.y, self.x, self.xy})

        self.cube.remove_coord("xy")
        self.assertEqual(set(self.cube.coords()), {self.y, self.x})

        self.cube.remove_coord("x")
        self.assertEqual(self.cube.coords(), [self.y])

        self.cube.remove_coord("y")
        self.assertEqual(self.cube.coords(), [])

    def test_immutable_dimcoord_dims(self):
        # Add DimCoord to dimension 1
        dims = [1]
        self.cube.add_dim_coord(self.x, dims)
        self.assertEqual(self.cube.coord_dims(self.x), (1,))

        # Change dims object
        dims[0] = 0
        # Check the cube is unchanged
        self.assertEqual(self.cube.coord_dims(self.x), (1,))

        # Check coord_dims cannot be changed
        dims = self.cube.coord_dims(self.x)
        with self.assertRaises(TypeError):
            dims[0] = 0

    def test_immutable_auxcoord_dims(self):
        # Add AuxCoord to dimensions (0, 1)
        dims = [0, 1]
        self.cube.add_aux_coord(self.xy, dims)
        self.assertEqual(self.cube.coord_dims(self.xy), (0, 1))

        # Change dims object
        dims[0] = 1
        dims[1] = 0
        # Check the cube is unchanged
        self.assertEqual(self.cube.coord_dims(self.xy), (0, 1))

        # Check coord_dims cannot be changed
        dims = self.cube.coord_dims(self.xy)
        with self.assertRaises(TypeError):
            dims[0] = 1


@tests.skip_data
class TestStockCubeStringRepresentations(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.realistic_4d()

    def test_4d_str(self):
        self.assertString(str(self.cube))

    def test_4d_repr(self):
        self.assertString(repr(self.cube))

    def test_3d_str(self):
        self.assertString(str(self.cube[0]))

    def test_3d_repr(self):
        self.assertString(repr(self.cube[0]))

    def test_2d_str(self):
        self.assertString(str(self.cube[0, 0]))

    def test_2d_repr(self):
        self.assertString(repr(self.cube[0, 0]))

    def test_1d_str(self):
        self.assertString(str(self.cube[0, 0, 0]))

    def test_1d_repr(self):
        self.assertString(repr(self.cube[0, 0, 0]))

    def test_0d_str(self):
        self.assertString(str(self.cube[0, 0, 0, 0]))

    def test_0d_repr(self):
        self.assertString(repr(self.cube[0, 0, 0, 0]))


@tests.skip_data
class TestCubeStringRepresentations(IrisDotTest):
    def setUp(self):
        path = tests.get_data_path(("PP", "simple_pp", "global.pp"))
        self.cube_2d = iris.load_cube(path)
        # Generate the unicode cube up here now it's used in two tests.
        unicode_str = chr(40960) + "abcd" + chr(1972)
        self.unicode_cube = iris.tests.stock.simple_1d()
        self.unicode_cube.attributes["source"] = unicode_str

    def test_dot_simple_pp(self):
        # Test dot output of a 2d cube loaded from pp.
        cube = self.cube_2d
        cube.attributes["my_attribute"] = "foobar"
        self.check_dot(cube, ("file_load", "global_pp.dot"))

        pt = cube.coord("time")
        # and with custom coord attributes
        pt.attributes["monty"] = "python"
        pt.attributes["brain"] = "hurts"
        self.check_dot(cube, ("file_load", "coord_attributes.dot"))

        del pt.attributes["monty"]
        del pt.attributes["brain"]
        del cube.attributes["my_attribute"]

    # TODO hybrid height and dot output - relatitionship links
    @tests.skip_data
    def test_dot_4d(self):
        cube = iris.tests.stock.realistic_4d()
        self.check_dot(cube, ("file_load", "4d_pp.dot"))

    @tests.skip_data
    def test_missing_coords(self):
        cube = iris.tests.stock.realistic_4d()
        cube.remove_coord("time")
        cube.remove_coord("model_level_number")
        self.assertString(
            repr(cube), ("cdm", "str_repr", "missing_coords_cube.repr.txt")
        )
        self.assertString(
            str(cube), ("cdm", "str_repr", "missing_coords_cube.str.txt")
        )

    @tests.skip_data
    def test_cubelist_string(self):
        cube_list = iris.cube.CubeList(
            [iris.tests.stock.realistic_4d(), iris.tests.stock.global_pp()]
        )
        self.assertString(
            str(cube_list), ("cdm", "str_repr", "cubelist.__str__.txt")
        )
        self.assertString(
            repr(cube_list), ("cdm", "str_repr", "cubelist.__repr__.txt")
        )

    def test_basic_0d_cube(self):
        self.assertString(
            repr(self.cube_2d[0, 0]),
            ("cdm", "str_repr", "0d_cube.__repr__.txt"),
        )
        self.assertString(
            str(self.cube_2d[0, 0]),
            ("cdm", "str_repr", "0d_cube.__unicode__.txt"),
        )
        self.assertString(
            str(self.cube_2d[0, 0]), ("cdm", "str_repr", "0d_cube.__str__.txt")
        )

    def test_similar_coord(self):
        cube = self.cube_2d.copy()

        lon = cube.coord("longitude")
        lon.attributes["flight"] = "218BX"
        lon.attributes["sensor_id"] = 808
        lon.attributes["status"] = 2
        lon2 = lon.copy()
        lon2.attributes["sensor_id"] = 810
        lon2.attributes["ref"] = "A8T-22"
        del lon2.attributes["status"]
        cube.add_aux_coord(lon2, [1])

        lat = cube.coord("latitude")
        lat2 = lat.copy()
        lat2.attributes["test"] = "True"
        cube.add_aux_coord(lat2, [0])

        self.assertString(
            str(cube), ("cdm", "str_repr", "similar.__str__.txt")
        )

    def test_cube_summary_cell_methods(self):

        cube = self.cube_2d.copy()

        # Create a list of values used to create cell methods
        test_values = (
            (
                ("mean",),
                ("longitude", "latitude"),
                ("6 minutes", "12 minutes"),
                ("This is a test comment",),
            ),
            (
                ("average",),
                ("longitude", "latitude"),
                ("6 minutes", "15 minutes"),
                ("This is another test comment", "This is another comment"),
            ),
            (("average",), ("longitude", "latitude"), (), ()),
            (
                ("percentile",),
                ("longitude",),
                ("6 minutes",),
                ("This is another test comment",),
            ),
        )

        for x in test_values:
            # Create a cell method
            cm = iris.coords.CellMethod(
                method=x[0][0], coords=x[1], intervals=x[2], comments=x[3]
            )
            cube.add_cell_method(cm)

        self.assertString(
            str(cube), ("cdm", "str_repr", "cell_methods.__str__.txt")
        )

    def test_cube_summary_alignment(self):
        # Test the cube summary dimension alignment and coord name clipping
        cube = iris.tests.stock.simple_1d()
        aux = iris.coords.AuxCoord(
            np.arange(11),
            long_name="This is a really, really, really, really long "
            "long_name that must be clipped because it is too long",
        )
        cube.add_aux_coord(aux, 0)
        aux = iris.coords.AuxCoord(
            np.arange(11), long_name="This is a short long_name"
        )
        cube.add_aux_coord(aux, 0)
        self.assertString(str(cube), ("cdm", "str_repr", "simple.__str__.txt"))

    def test_unicode_attribute(self):
        self.assertString(
            str(self.unicode_cube),
            ("cdm", "str_repr", "unicode_attribute.__unicode__.txt"),
        )


@tests.skip_data
class TestValidity(tests.IrisTest):
    def setUp(self):
        self.cube_2d = iris.load_cube(
            tests.get_data_path(("PP", "simple_pp", "global.pp"))
        )

    def test_wrong_length_vector_coord(self):
        wobble = iris.coords.DimCoord(
            points=[1, 2], long_name="wobble", units="1"
        )
        with self.assertRaises(ValueError):
            self.cube_2d.add_aux_coord(wobble, 0)

    def test_invalid_dimension_vector_coord(self):
        wobble = iris.coords.DimCoord(
            points=[1, 2], long_name="wobble", units="1"
        )
        with self.assertRaises(ValueError):
            self.cube_2d.add_dim_coord(wobble, 99)


class TestQueryCoord(tests.IrisTest):
    def setUp(self):
        self.t = iris.tests.stock.simple_2d_w_multidim_and_scalars()

    def test_name(self):
        coords = self.t.coords("dim1")
        self.assertEqual([coord.name() for coord in coords], ["dim1"])

        coords = self.t.coords("dim2")
        self.assertEqual([coord.name() for coord in coords], ["dim2"])

        coords = self.t.coords("an_other")
        self.assertEqual([coord.name() for coord in coords], ["an_other"])

        coords = self.t.coords("air_temperature")
        self.assertEqual(
            [coord.name() for coord in coords], ["air_temperature"]
        )

        coords = self.t.coords("wibble")
        self.assertEqual(coords, [])

    def test_long_name(self):
        # Both standard_name and long_name defined
        coords = self.t.coords(long_name="custom long name")
        # coord.name() returns standard_name if available
        self.assertEqual(
            [coord.name() for coord in coords], ["air_temperature"]
        )

    def test_standard_name(self):
        # Both standard_name and long_name defined
        coords = self.t.coords(standard_name="custom long name")
        self.assertEqual([coord.name() for coord in coords], [])
        coords = self.t.coords(standard_name="air_temperature")
        self.assertEqual(
            [coord.name() for coord in coords], ["air_temperature"]
        )

    def test_var_name(self):
        coords = self.t.coords(var_name="custom_var_name")
        # Matching coord in test cube has a standard_name of 'air_temperature'.
        self.assertEqual(
            [coord.name() for coord in coords], ["air_temperature"]
        )

    def test_axis(self):
        cube = self.t.copy()
        cube.coord("dim1").rename("latitude")
        cube.coord("dim2").rename("longitude")

        coords = cube.coords(axis="y")
        self.assertEqual([coord.name() for coord in coords], ["latitude"])

        coords = cube.coords(axis="x")
        self.assertEqual([coord.name() for coord in coords], ["longitude"])

        # Renaming shoudn't be enough
        cube.coord("an_other").rename("time")
        coords = cube.coords(axis="t")
        self.assertEqual([coord.name() for coord in coords], [])
        # Change units to "hours since ..." as it's the presence of a
        # time unit that identifies a time axis.
        cube.coord("time").units = "hours since 1970-01-01 00:00:00"
        coords = cube.coords(axis="t")
        self.assertEqual([coord.name() for coord in coords], ["time"])

        coords = cube.coords(axis="z")
        self.assertEqual(coords, [])

    def test_contains_dimension(self):
        coords = self.t.coords(contains_dimension=0)
        self.assertEqual(
            [coord.name() for coord in coords], ["dim1", "my_multi_dim_coord"]
        )

        coords = self.t.coords(contains_dimension=1)
        self.assertEqual(
            [coord.name() for coord in coords], ["dim2", "my_multi_dim_coord"]
        )

        coords = self.t.coords(contains_dimension=2)
        self.assertEqual(coords, [])

    def test_dimensions(self):
        coords = self.t.coords(dimensions=0)
        self.assertEqual([coord.name() for coord in coords], ["dim1"])

        coords = self.t.coords(dimensions=1)
        self.assertEqual([coord.name() for coord in coords], ["dim2"])

        # find all coordinates which do not describe a dimension
        coords = self.t.coords(dimensions=[])
        self.assertEqual(
            [coord.name() for coord in coords], ["air_temperature", "an_other"]
        )

        coords = self.t.coords(dimensions=2)
        self.assertEqual(coords, [])

        coords = self.t.coords(dimensions=[0, 1])
        self.assertEqual(
            [coord.name() for coord in coords], ["my_multi_dim_coord"]
        )

    def test_coord_dim_coords_keyword(self):
        coords = self.t.coords(dim_coords=True)
        self.assertEqual(
            set([coord.name() for coord in coords]), {"dim1", "dim2"}
        )

        coords = self.t.coords(dim_coords=False)
        self.assertEqual(
            set([coord.name() for coord in coords]),
            {"an_other", "my_multi_dim_coord", "air_temperature"},
        )

    def test_coords_empty(self):
        coords = self.t.coords()
        self.assertEqual(
            set([coord.name() for coord in coords]),
            {
                "dim1",
                "dim2",
                "an_other",
                "my_multi_dim_coord",
                "air_temperature",
            },
        )

    def test_coord(self):
        coords = self.t.coords(self.t.coord("dim1"))
        self.assertEqual([coord.name() for coord in coords], ["dim1"])
        # check for metadata look-up by modifying points
        coord = self.t.coord("dim1").copy()
        coord.points = np.arange(5) * 1.23
        coords = self.t.coords(coord)
        self.assertEqual([coord.name() for coord in coords], ["dim1"])

    def test_str_repr(self):
        # TODO consolidate with the TestCubeStringRepresentations class
        self.assertString(
            str(self.t), ("cdm", "str_repr", "multi_dim_coord.__str__.txt")
        )
        self.assertString(
            repr(self.t), ("cdm", "str_repr", "multi_dim_coord.__repr__.txt")
        )


class TestCube2d(tests.IrisTest):
    def setUp(self):
        self.t = iris.tests.stock.simple_2d_w_multidim_and_scalars()
        self.t.remove_coord("air_temperature")


class Test2dIndexing(TestCube2d):
    def test_indexing_of_0d_cube(self):
        c = self.t[0, 0]
        self.assertRaises(IndexError, c.__getitem__, (slice(None, None),))

    def test_cube_indexing_0d(self):
        self.assertCML(
            [self.t[0, 0]], ("cube_slice", "2d_to_0d_cube_slice.cml")
        )

    def test_cube_indexing_1d(self):
        self.assertCML(
            [self.t[0, 0:]], ("cube_slice", "2d_to_1d_cube_slice.cml")
        )

    def test_cube_indexing_1d_multi_slice(self):
        self.assertCML(
            [self.t[0, (0, 1)]],
            ("cube_slice", "2d_to_1d_cube_multi_slice.cml"),
        )
        self.assertCML(
            [self.t[0, np.array([0, 1])]],
            ("cube_slice", "2d_to_1d_cube_multi_slice.cml"),
        )

    def test_cube_indexing_1d_multi_slice2(self):
        self.assertCML(
            [self.t[(0, 2), (0, 1, 3)]],
            ("cube_slice", "2d_to_1d_cube_multi_slice2.cml"),
        )
        self.assertCML(
            [self.t[np.array([0, 2]), (0, 1, 3)]],
            ("cube_slice", "2d_to_1d_cube_multi_slice2.cml"),
        )
        self.assertCML(
            [self.t[np.array([0, 2]), np.array([0, 1, 3])]],
            ("cube_slice", "2d_to_1d_cube_multi_slice2.cml"),
        )

    def test_cube_indexing_1d_multi_slice3(self):
        self.assertCML(
            [self.t[(0, 2), :]],
            ("cube_slice", "2d_to_1d_cube_multi_slice3.cml"),
        )
        self.assertCML(
            [self.t[np.array([0, 2]), :]],
            ("cube_slice", "2d_to_1d_cube_multi_slice3.cml"),
        )

    def test_cube_indexing_no_change(self):
        self.assertCML([self.t[0:, 0:]], ("cube_slice", "2d_orig.cml"))

    def test_cube_indexing_reverse_coords(self):
        self.assertCML(
            [self.t[::-1, ::-1]], ("cube_slice", "2d_to_2d_revesed.cml")
        )

    def test_cube_indexing_no_residual_change(self):
        self.t[0:3]
        self.assertCML([self.t], ("cube_slice", "2d_orig.cml"))

    def test_overspecified(self):
        self.assertRaises(IndexError, self.t.__getitem__, (0, 0, Ellipsis, 0))
        self.assertRaises(IndexError, self.t.__getitem__, (0, 0, 0))

    def test_ellipsis(self):
        self.assertCML([self.t[Ellipsis]], ("cube_slice", "2d_orig.cml"))
        self.assertCML([self.t[:, :, :]], ("cube_slice", "2d_orig.cml"))
        self.assertCML(
            [self.t[Ellipsis, Ellipsis]], ("cube_slice", "2d_orig.cml")
        )
        self.assertCML(
            [self.t[Ellipsis, Ellipsis, Ellipsis]],
            ("cube_slice", "2d_orig.cml"),
        )

        self.assertCML(
            [self.t[Ellipsis, 0, 0]], ("cube_slice", "2d_to_0d_cube_slice.cml")
        )
        self.assertCML(
            [self.t[0, Ellipsis, 0]], ("cube_slice", "2d_to_0d_cube_slice.cml")
        )
        self.assertCML(
            [self.t[0, 0, Ellipsis]], ("cube_slice", "2d_to_0d_cube_slice.cml")
        )

        self.assertCML(
            [self.t[Ellipsis, (0, 2), :]],
            ("cube_slice", "2d_to_1d_cube_multi_slice3.cml"),
        )
        self.assertCML(
            [self.t[(0, 2), Ellipsis, :]],
            ("cube_slice", "2d_to_1d_cube_multi_slice3.cml"),
        )
        self.assertCML(
            [self.t[(0, 2), :, Ellipsis]],
            ("cube_slice", "2d_to_1d_cube_multi_slice3.cml"),
        )


class TestIteration(TestCube2d):
    def test_cube_iteration(self):
        with self.assertRaises(TypeError):
            for subcube in self.t:
                pass

    def test_not_iterable(self):
        self.assertFalse(isinstance(self.t, collections.Iterable))


class Test2dSlicing(TestCube2d):
    def test_cube_slice_all_dimensions(self):
        for cube in self.t.slices(["dim1", "dim2"]):
            self.assertCML(cube, ("cube_slice", "2d_orig.cml"))

    def test_cube_slice_with_transpose(self):
        for cube in self.t.slices(["dim2", "dim1"]):
            self.assertCML(cube, ("cube_slice", "2d_transposed.cml"))

    def test_cube_slice_without_transpose(self):
        for cube in self.t.slices(["dim2", "dim1"], ordered=False):
            self.assertCML(cube, ("cube_slice", "2d_orig.cml"))

    def test_cube_slice_1dimension(self):
        # Result came from the equivalent test test_cube_indexing_1d which
        # does self.t[0, 0:]
        slices = [res for res in self.t.slices(["dim2"])]
        self.assertCML(slices[0], ("cube_slice", "2d_to_1d_cube_slice.cml"))

    def test_cube_slice_zero_len_slice(self):
        self.assertRaises(IndexError, self.t.__getitem__, (slice(0, 0)))

    def test_cube_slice_with_non_existant_coords(self):
        with self.assertRaises(iris.exceptions.CoordinateNotFoundError):
            self.t.slices(["dim2", "dim1", "doesnt exist"])

    def test_cube_extract_coord_with_non_describing_coordinates(self):
        with self.assertRaises(ValueError):
            self.t.slices(["an_other"])


class Test2dSlicing_ByDim(TestCube2d):
    def test_cube_slice_all_dimensions(self):
        for cube in self.t.slices([0, 1]):
            self.assertCML(cube, ("cube_slice", "2d_orig.cml"))

    def test_cube_slice_with_transpose(self):
        for cube in self.t.slices([1, 0]):
            self.assertCML(cube, ("cube_slice", "2d_transposed.cml"))

    def test_cube_slice_without_transpose(self):
        for cube in self.t.slices([1, 0], ordered=False):
            self.assertCML(cube, ("cube_slice", "2d_orig.cml"))

    def test_cube_slice_1dimension(self):
        # Result came from the equivalent test test_cube_indexing_1d which
        # does self.t[0, 0:]
        slices = [res for res in self.t.slices([1])]
        self.assertCML(slices[0], ("cube_slice", "2d_to_1d_cube_slice.cml"))

    def test_cube_slice_nodimension(self):
        slices = [res for res in self.t.slices([])]
        self.assertCML(slices[0], ("cube_slice", "2d_to_0d_cube_slice.cml"))

    def test_cube_slice_with_non_existant_dims(self):
        with self.assertRaises(IndexError):
            self.t.slices([1, 0, 2])

    def test_cube_slice_duplicate_dimensions(self):
        with self.assertRaises(ValueError):
            self.t.slices([1, 1])


class Test2dSlicing_ByMix(TestCube2d):
    def test_cube_slice_all_dimensions(self):
        for cube in self.t.slices([0, "dim2"]):
            self.assertCML(cube, ("cube_slice", "2d_orig.cml"))

    def test_cube_slice_with_transpose(self):
        for cube in self.t.slices(["dim2", 0]):
            self.assertCML(cube, ("cube_slice", "2d_transposed.cml"))

    def test_cube_slice_with_non_existant_dims(self):
        with self.assertRaises(ValueError):
            self.t.slices([1, 0, "an_other"])


class Test2dExtraction(TestCube2d):
    def test_cube_extract_0d(self):
        # Extract the first value from each of the coords in the cube
        # this result is shared with the self.t[0, 0] test
        self.assertCML(
            [
                self.t.extract(
                    iris.Constraint(dim1=3.0, dim2=iris.coords.Cell(0, (0, 1)))
                )
            ],
            ("cube_slice", "2d_to_0d_cube_slice.cml"),
        )

    def test_cube_extract_1d(self):
        # Extract the first value from the second coord in the cube
        # this result is shared with the self.t[0, 0:] test
        self.assertCML(
            [self.t.extract(iris.Constraint(dim1=3.0))],
            ("cube_slice", "2d_to_1d_cube_slice.cml"),
        )

    def test_cube_extract_2d(self):
        # Do nothing - return the original
        self.assertCML(
            [self.t.extract(iris.Constraint())], ("cube_slice", "2d_orig.cml")
        )

    def test_cube_extract_coord_which_does_not_exist(self):
        self.assertEqual(
            self.t.extract(iris.Constraint(doesnt_exist=8.1)), None
        )

    def test_cube_extract_coord_with_non_existant_values(self):
        self.assertEqual(self.t.extract(iris.Constraint(dim1=8)), None)


class Test2dExtractionByCoord(TestCube2d):
    def test_cube_extract_by_coord_advanced(self):
        # This test reverses the coordinate in the cube and also takes a subset of the original coordinate
        points = np.array([9, 8, 7, 5, 4, 3, 2, 1, 0], dtype=np.int32)
        bounds = np.array(
            [
                [18, 19],
                [16, 17],
                [14, 15],
                [10, 11],
                [8, 9],
                [6, 7],
                [4, 5],
                [2, 3],
                [0, 1],
            ],
            dtype=np.int32,
        )
        c = iris.coords.DimCoord(
            points, long_name="dim2", units="meters", bounds=bounds
        )
        self.assertCML(
            self.t.subset(c), ("cube_slice", "2d_intersect_and_reverse.cml")
        )


@tests.skip_data
class TestCubeExtract(tests.IrisTest):
    def setUp(self):
        self.single_cube = iris.load_cube(
            tests.get_data_path(("PP", "globClim1", "theta.pp")),
            "air_potential_temperature",
        )

    def test_simple(self):
        constraint = iris.Constraint(latitude=10)
        cube = self.single_cube.extract(constraint)
        self.assertCML(cube, ("cdm", "extract", "lat_eq_10.cml"))
        constraint = iris.Constraint(latitude=lambda c: c > 10)
        self.assertCML(
            self.single_cube.extract(constraint),
            ("cdm", "extract", "lat_gt_10.cml"),
        )

    def test_combined(self):
        constraint = iris.Constraint(
            latitude=lambda c: c > 10, longitude=lambda c: c >= 10
        )

        self.assertCML(
            self.single_cube.extract(constraint),
            ("cdm", "extract", "lat_gt_10_and_lon_ge_10.cml"),
        )

    def test_no_results(self):
        constraint = iris.Constraint(latitude=lambda c: c > 1000000)
        self.assertEqual(self.single_cube.extract(constraint), None)


class TestCubeAPI(TestCube2d):
    def test_getting_standard_name(self):
        self.assertEqual(self.t.name(), "test 2d dimensional cube")

    def test_rename(self):
        self.t.rename("foo")
        self.assertEqual(self.t.name(), "foo")

    def test_var_name(self):
        self.t.var_name = None
        self.assertEqual(self.t.var_name, None)
        self.t.var_name = "bar"
        self.assertEqual(self.t.var_name, "bar")

    def test_default_name(self):
        self.t.long_name = ""
        self.assertEqual(self.t.name(), "unknown")
        self.assertEqual(self.t.name("a_default"), "a_default")

    def test_stash_name(self):
        self.t.long_name = ""
        self.t.attributes["STASH"] = iris.fileformats.pp.STASH(1, 2, 3)
        self.assertEqual(self.t.name(), "m01s02i003")

    def test_name_and_var_name(self):
        # Assign only var_name.
        self.t.standard_name = None
        self.t.long_name = None
        self.t.var_name = "foo"
        # name() should return var_name if standard_name and
        # long_name are None.
        self.assertEqual(self.t.name(), "foo")

    def test_rename_and_var_name(self):
        self.t.var_name = "bar"
        self.t.rename("foo")
        # Rename should clear var_name.
        self.assertIsNone(self.t.var_name)

    def test_setting_invalid_var_name(self):
        # Name with whitespace should raise an exception.
        with self.assertRaises(ValueError):
            self.t.var_name = "foo bar"

    def test_setting_empty_var_name(self):
        # Empty string should raise an exception.
        with self.assertRaises(ValueError):
            self.t.var_name = ""

    def test_getting_units(self):
        self.assertEqual(self.t.units, cf_units.Unit("meters"))

    def test_setting_units(self):
        self.assertEqual(self.t.units, cf_units.Unit("meters"))
        self.t.units = "kelvin"
        self.assertEqual(self.t.units, cf_units.Unit("kelvin"))

    def test_clearing_units(self):
        self.t.units = None
        self.assertEqual(str(self.t.units), "unknown")

    def test_convert_units(self):
        # Set to 'volt'
        self.t.units = cf_units.Unit("volt")
        data = self.t.data.copy()
        # Change to 'kV' - data should be scaled automatically.
        self.t.convert_units("kV")
        self.assertEqual(str(self.t.units), "kV")
        self.assertArrayAlmostEqual(self.t.data, data / 1000.0)

    def test_coords_are_copies(self):
        self.assertIsNot(self.t.coord("dim1"), self.t.copy().coord("dim1"))

    def test_metadata_nop(self):
        self.t.metadata = self.t.metadata
        self.assertIsNone(self.t.standard_name)
        self.assertEqual(self.t.long_name, "test 2d dimensional cube")
        self.assertIsNone(self.t.var_name)
        self.assertEqual(self.t.units, "meters")
        self.assertEqual(self.t.attributes, {})
        self.assertEqual(self.t.cell_methods, ())

    def test_metadata_tuple(self):
        metadata = ("air_pressure", "foo", "bar", "", {"random": "12"}, ())
        self.t.metadata = metadata
        self.assertEqual(self.t.standard_name, "air_pressure")
        self.assertEqual(self.t.long_name, "foo")
        self.assertEqual(self.t.var_name, "bar")
        self.assertEqual(self.t.units, "")
        self.assertEqual(self.t.attributes, metadata[4])
        self.assertIsNot(self.t.attributes, metadata[4])
        self.assertEqual(self.t.cell_methods, ())

    def test_metadata_dict(self):
        metadata = {
            "standard_name": "air_pressure",
            "long_name": "foo",
            "var_name": "bar",
            "units": "",
            "attributes": {"random": "12"},
            "cell_methods": (),
        }
        self.t.metadata = metadata
        self.assertEqual(self.t.standard_name, "air_pressure")
        self.assertEqual(self.t.long_name, "foo")
        self.assertEqual(self.t.var_name, "bar")
        self.assertEqual(self.t.units, "")
        self.assertEqual(self.t.attributes, metadata["attributes"])
        self.assertIsNot(self.t.attributes, metadata["attributes"])
        self.assertEqual(self.t.cell_methods, ())

    def test_metadata_attrs(self):
        class Metadata:
            pass

        metadata = Metadata()
        metadata.standard_name = "air_pressure"
        metadata.long_name = "foo"
        metadata.var_name = "bar"
        metadata.units = ""
        metadata.attributes = {"random": "12"}
        metadata.cell_methods = ()
        metadata.cell_measures_and_dims = []
        self.t.metadata = metadata
        self.assertEqual(self.t.standard_name, "air_pressure")
        self.assertEqual(self.t.long_name, "foo")
        self.assertEqual(self.t.var_name, "bar")
        self.assertEqual(self.t.units, "")
        self.assertEqual(self.t.attributes, metadata.attributes)
        self.assertIsNot(self.t.attributes, metadata.attributes)
        self.assertEqual(self.t.cell_methods, ())
        self.assertEqual(self.t._cell_measures_and_dims, [])

    def test_metadata_fail(self):
        with self.assertRaises(TypeError):
            self.t.metadata = (
                "air_pressure",
                "foo",
                "bar",
                "",
                {"random": "12"},
            )
        with self.assertRaises(TypeError):
            self.t.metadata = (
                "air_pressure",
                "foo",
                "bar",
                "",
                {"random": "12"},
                (),
                [],
                (),
                (),
            )
        with self.assertRaises(TypeError):

            class Metadata:
                pass

            metadata = Metadata()
            metadata.standard_name = "air_pressure"
            metadata.long_name = "foo"
            metadata.var_name = "bar"
            metadata.units = ""
            metadata.attributes = {"random": "12"}
            self.t.metadata = metadata


class TestCubeEquality(TestCube2d):
    def test_simple_equality(self):
        self.assertEqual(self.t, self.t.copy())

    def test_data_inequality(self):
        self.assertNotEqual(self.t, self.t + 1)

    def test_coords_inequality(self):
        r = self.t.copy()
        r.remove_coord(r.coord("an_other"))
        self.assertNotEqual(self.t, r)

    def test_attributes_inequality(self):
        r = self.t.copy()
        r.attributes["new_thing"] = None
        self.assertNotEqual(self.t, r)

    def test_array_attributes(self):
        r = self.t.copy()
        r.attributes["things"] = np.arange(3)
        s = r.copy()
        self.assertEqual(s, r)

        s.attributes["things"] = np.arange(2)
        self.assertNotEqual(s, r)

        del s.attributes["things"]
        self.assertNotEqual(s, r)

    def test_cell_methods_inequality(self):
        r = self.t.copy()
        r.add_cell_method(iris.coords.CellMethod("mean"))
        self.assertNotEqual(self.t, r)

    def test_not_compatible(self):
        r = self.t.copy()
        self.assertTrue(self.t.is_compatible(r))
        # The following changes should make the cubes incompatible.
        # Different units.
        r.units = "kelvin"
        self.assertFalse(self.t.is_compatible(r))
        # Different cell_methods.
        r = self.t.copy()
        r.add_cell_method(iris.coords.CellMethod("mean", coords="dim1"))
        self.assertFalse(self.t.is_compatible(r))
        # Different attributes.
        r = self.t.copy()
        self.t.attributes["source"] = "bob"
        r.attributes["source"] = "alice"
        self.assertFalse(self.t.is_compatible(r))

    def test_compatible(self):
        r = self.t.copy()
        self.assertTrue(self.t.is_compatible(r))
        # The following changes should not affect compatibility.
        # Different non-common attributes.
        self.t.attributes["source"] = "bob"
        r.attributes["origin"] = "alice"
        self.assertTrue(self.t.is_compatible(r))
        # Different coordinates.
        r.remove_coord("dim1")
        self.assertTrue(self.t.is_compatible(r))
        # Different data.
        r.data = np.zeros(r.shape)
        self.assertTrue(self.t.is_compatible(r))
        # Different var_names (but equal name()).
        r.var_name = "foo"
        self.assertTrue(self.t.is_compatible(r))

    def test_is_compatible_ignore(self):
        r = self.t.copy()
        self.assertTrue(self.t.is_compatible(r))
        # Different histories.
        self.t.attributes["history"] = "One history."
        r.attributes["history"] = "An alternative history."
        self.assertFalse(self.t.is_compatible(r))
        # Use ignore keyword.
        self.assertTrue(self.t.is_compatible(r, ignore="history"))
        self.assertTrue(self.t.is_compatible(r, ignore=("history",)))
        self.assertTrue(self.t.is_compatible(r, ignore=r.attributes))

    def test_is_compatible_metadata(self):
        metadata = self.t.metadata
        self.assertTrue(self.t.is_compatible(metadata))


@tests.skip_data
class TestDataManagerIndexing(TestCube2d):
    def setUp(self):
        self.cube = iris.load_cube(
            tests.get_data_path(("PP", "aPProt1", "rotatedMHtimecube.pp"))
        )

    def assert_is_lazy(self, cube):
        self.assertTrue(cube.has_lazy_data())

    def assert_is_not_lazy(self, cube):
        self.assertFalse(cube.has_lazy_data())

    def test_slices(self):
        lat_cube = next(self.cube.slices(["grid_latitude"]))
        self.assert_is_lazy(lat_cube)
        self.assert_is_lazy(self.cube)

    def test_cube_empty_indexing(self):
        test_filename = ("cube_slice", "real_empty_data_indexing.cml")
        r = self.cube[:5, ::-1][3]
        rshape = r.shape

        # Make sure we still have deferred data.
        self.assert_is_lazy(r)
        # check the CML of this result
        self.assertCML(r, test_filename)
        # The CML was checked, meaning the data must have been loaded.
        # Check that the cube no longer has deferred data.
        self.assert_is_not_lazy(r)

        r_data = r.data

        # finally, load the data before indexing and check that it generates the same result
        c = self.cube
        c.data
        c = c[:5, ::-1][3]
        self.assertCML(c, test_filename)

        self.assertEqual(rshape, c.shape)

        np.testing.assert_array_equal(r_data, c.data)

    def test_real_data_cube_indexing(self):
        cube = self.cube[(0, 4, 5, 2), 0, 0]
        self.assertCML(
            cube, ("cube_slice", "real_data_dual_tuple_indexing1.cml")
        )

        cube = self.cube[0, (0, 4, 5, 2), (3, 5, 5)]
        self.assertCML(
            cube, ("cube_slice", "real_data_dual_tuple_indexing2.cml")
        )

        cube = self.cube[(0, 4, 5, 2), 0, (3, 5, 5)]
        self.assertCML(
            cube, ("cube_slice", "real_data_dual_tuple_indexing3.cml")
        )

        self.assertRaises(
            IndexError,
            self.cube.__getitem__,
            ((0, 4, 5, 2), (3, 5, 5), 0, 0, 4),
        )
        self.assertRaises(
            IndexError,
            self.cube.__getitem__,
            (Ellipsis, Ellipsis, Ellipsis, Ellipsis, Ellipsis, Ellipsis),
        )

    def test_fancy_indexing_bool_array(self):
        cube = self.cube
        cube.data = np.ma.masked_array(cube.data, mask=cube.data > 100000)
        r = cube[:, cube.coord("grid_latitude").points > 1]
        self.assertEqual(r.shape, (10, 218, 720))
        data = cube.data[:, self.cube.coord("grid_latitude").points > 1, :]
        np.testing.assert_array_equal(data, r.data)
        np.testing.assert_array_equal(data.mask, r.data.mask)


class TestCubeCollapsed(tests.IrisTest):
    def partial_compare(self, dual, single):
        result = iris.analysis._dimensional_metadata_comparison(dual, single)
        self.assertEqual(len(result["not_equal"]), 0)
        self.assertEqual(
            dual.name(),
            single.name(),
            "dual and single stage standard_names differ",
        )
        self.assertEqual(
            dual.units, single.units, "dual and single stage units differ"
        )
        self.assertEqual(
            dual.shape, single.shape, "dual and single stage shape differ"
        )

    def collapse_test_common(self, cube, a_name, b_name, *args, **kwargs):
        # preserve filenames from before the introduction of "grid_" in rotated coord names.
        a_filename = a_name.replace("grid_", "")
        b_filename = b_name.replace("grid_", "")

        # compare dual and single stage collapsing
        dual_stage = cube.collapsed(a_name, iris.analysis.MEAN)
        dual_stage = dual_stage.collapsed(b_name, iris.analysis.MEAN)
        # np.ma.average doesn't apply type promotion rules in some versions,
        # and instead makes the result type float64. To ignore that case we
        # fix up the dtype here if it is promotable from cube.dtype. We still
        # want to catch cases where there is a loss of precision however.
        if dual_stage.dtype > cube.dtype:
            data = dual_stage.data.astype(cube.dtype)
            dual_stage.data = data
        self.assertCMLApproxData(
            dual_stage,
            (
                "cube_collapsed",
                "%s_%s_dual_stage.cml" % (a_filename, b_filename),
            ),
            *args,
            **kwargs,
        )

        single_stage = cube.collapsed([a_name, b_name], iris.analysis.MEAN)
        if single_stage.dtype > cube.dtype:
            data = single_stage.data.astype(cube.dtype)
            single_stage.data = data
        self.assertCMLApproxData(
            single_stage,
            (
                "cube_collapsed",
                "%s_%s_single_stage.cml" % (a_filename, b_filename),
            ),
            *args,
            **kwargs,
        )

        # Compare the cube bits that should match
        self.partial_compare(dual_stage, single_stage)

    @tests.skip_data
    def test_multi_d(self):
        cube = iris.tests.stock.realistic_4d()

        # TODO: Re-instate surface_altitude & hybrid-height once we're
        # using the post-CF test results.
        cube.remove_aux_factory(cube.aux_factories[0])
        cube.remove_coord("surface_altitude")

        self.assertCML(cube, ("cube_collapsed", "original.cml"))

        # Compare 2-stage collapsing with a single stage collapse
        # over 2 Coords.
        self.collapse_test_common(
            cube, "grid_latitude", "grid_longitude", rtol=1e-05
        )
        self.collapse_test_common(
            cube, "grid_longitude", "grid_latitude", rtol=1e-05
        )

        self.collapse_test_common(cube, "time", "grid_latitude", rtol=1e-05)
        self.collapse_test_common(cube, "grid_latitude", "time", rtol=1e-05)

        self.collapse_test_common(cube, "time", "grid_longitude", rtol=1e-05)
        self.collapse_test_common(cube, "grid_longitude", "time", rtol=1e-05)

        self.collapse_test_common(
            cube, "grid_latitude", "model_level_number", rtol=5e-04
        )
        self.collapse_test_common(
            cube, "model_level_number", "grid_latitude", rtol=5e-04
        )

        self.collapse_test_common(
            cube, "grid_longitude", "model_level_number", rtol=5e-04
        )
        self.collapse_test_common(
            cube, "model_level_number", "grid_longitude", rtol=5e-04
        )

        self.collapse_test_common(
            cube, "time", "model_level_number", rtol=5e-04
        )
        self.collapse_test_common(
            cube, "model_level_number", "time", rtol=5e-04
        )

        self.collapse_test_common(
            cube, "model_level_number", "time", rtol=5e-04
        )
        self.collapse_test_common(
            cube, "time", "model_level_number", rtol=5e-04
        )

        # Collapse 3 things at once.
        triple_collapse = cube.collapsed(
            ["model_level_number", "time", "grid_longitude"],
            iris.analysis.MEAN,
        )
        self.assertCMLApproxData(
            triple_collapse,
            ("cube_collapsed", ("triple_collapse_ml_pt_" "lon.cml")),
            rtol=5e-04,
        )

        triple_collapse = cube.collapsed(
            ["grid_latitude", "model_level_number", "time"], iris.analysis.MEAN
        )
        self.assertCMLApproxData(
            triple_collapse,
            ("cube_collapsed", ("triple_collapse_lat_ml" "_pt.cml")),
            rtol=0.05,
        )
        # KNOWN PROBLEM: the previous 'rtol' is very large.
        # Numpy 1.10 and 1.11 give significantly different results here.
        # This may relate to known problems with summing over large arrays,
        # which were largely fixed in numpy 1.9 but still occur in some cases,
        # as-of numpy 1.11.

        # Ensure no side effects
        self.assertCML(cube, ("cube_collapsed", "original.cml"))


@tests.skip_data
class TestTrimAttributes(tests.IrisTest):
    def test_non_string_attributes(self):
        cube = iris.tests.stock.realistic_4d()
        attrib_key = "gorf"
        attrib_val = 23
        cube.attributes[attrib_key] = attrib_val

        summary = cube.summary()  # Get the cube summary

        # Check through the lines of the summary to see that our attribute is there
        attrib_re = re.compile("%s.*?%s" % (attrib_key, attrib_val))

        for line in summary.split("\n"):
            result = re.match(attrib_re, line.strip())
            if result:
                break
        else:  # No match found for our attribute
            self.fail("Attribute not found in summary output of cube.")


@tests.skip_data
class TestMaskedData(tests.IrisTest, pp.PPTest):
    def _load_3d_cube(self):
        # This 3D data set has a missing a slice with SOME missing values.
        # The missing data is in the pressure = 1000 hPa, forcast_period = 0,
        # time = 1970-02-11 16:00:00 slice.
        return iris.load_cube(
            tests.get_data_path(["PP", "mdi_handmade_small", "*.pp"])
        )

    def test_complete_field(self):
        # This pp field has no missing data values
        cube = iris.load_cube(
            tests.get_data_path(
                ["PP", "mdi_handmade_small", "mdi_test_1000_3.pp"]
            )
        )

        self.assertIsInstance(cube.data, np.ndarray)

    def test_masked_field(self):
        # This pp field has some missing data values
        cube = iris.load_cube(
            tests.get_data_path(
                ["PP", "mdi_handmade_small", "mdi_test_1000_0.pp"]
            )
        )
        self.assertIsInstance(cube.data, ma.core.MaskedArray)

    def test_missing_file(self):
        cube = self._load_3d_cube()
        self.assertIsInstance(cube.data, ma.core.MaskedArray)
        self.assertCML(cube, ("cdm", "masked_cube.cml"))

    def test_slicing(self):
        cube = self._load_3d_cube()

        # Test the slicing before deferred loading
        full_slice = cube[3]
        partial_slice = cube[0]
        self.assertIsInstance(full_slice.data, np.ndarray)
        self.assertIsInstance(partial_slice.data, ma.core.MaskedArray)
        self.assertEqual(ma.count_masked(partial_slice.data), 25)

        # Test the slicing is consistent after deferred loading
        full_slice = cube[3]
        partial_slice = cube[0]
        self.assertIsInstance(full_slice.data, np.ndarray)
        self.assertIsInstance(partial_slice.data, ma.core.MaskedArray)
        self.assertEqual(ma.count_masked(partial_slice.data), 25)

    def test_save_and_merge(self):
        cube = self._load_3d_cube()
        dtype = cube.dtype
        fill_value = 123456

        # extract the 2d field that has SOME missing values
        masked_slice = cube[0]
        masked_slice.data.fill_value = fill_value

        # test saving masked data
        reference_txt_path = tests.get_result_path(
            ("cdm", "masked_save_pp.txt")
        )
        with self.cube_save_test(
            reference_txt_path, reference_cubes=masked_slice
        ) as temp_pp_path:
            iris.save(masked_slice, temp_pp_path)

            # test merge keeps the mdi we just saved
            cube1 = iris.load_cube(temp_pp_path)
            self.assertEqual(cube1.dtype, dtype)

            cube2 = cube1.copy()
            # make cube1 and cube2 differ on a scalar coord, to make them mergeable into a 3d cube
            cube2.coord("pressure").points = [1001.0]
            merged_cubes = iris.cube.CubeList([cube1, cube2]).merge()
            self.assertEqual(
                len(merged_cubes), 1, "expected a single merged cube"
            )
            merged_cube = merged_cubes[0]
            self.assertEqual(merged_cube.dtype, dtype)
            # Check that the original masked-array fill-value is *ignored*.
            self.assertArrayAllClose(merged_cube.data.fill_value, -1e30)


@tests.skip_data
class TestConversionToCoordList(tests.IrisTest):
    def test_coord_conversion(self):
        cube = iris.tests.stock.realistic_4d()

        # Single string
        self.assertEqual(len(cube._as_list_of_coords("grid_longitude")), 1)

        # List of string and unicode
        self.assertEqual(
            len(cube._as_list_of_coords(["grid_longitude", "grid_latitude"])),
            2,
        )

        # Coord object(s)
        lat = cube.coords("grid_latitude")[0]
        lon = cube.coords("grid_longitude")[0]
        self.assertEqual(len(cube._as_list_of_coords(lat)), 1)
        self.assertEqual(len(cube._as_list_of_coords([lat, lon])), 2)

        # Mix of string-like and coord
        self.assertEqual(
            len(cube._as_list_of_coords(["grid_latitude", lon])), 2
        )

        # Empty list
        self.assertEqual(len(cube._as_list_of_coords([])), 0)

        # Invalid coords
        invalid_choices = [
            iris.analysis.MEAN,  # Caused by mixing up argument order in call to cube.collasped for example
            None,
            ["grid_latitude", None],
            [lat, None],
        ]

        for coords in invalid_choices:
            with self.assertRaises(TypeError):
                cube._as_list_of_coords(coords)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :func:`iris.analysis.maths.add` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import operator

from iris.analysis.maths import add
from iris.tests.unit.analysis.maths import (
    CubeArithmeticBroadcastingTestMixin,
    CubeArithmeticCoordsTest,
    CubeArithmeticMaskedConstantTestMixin,
    CubeArithmeticMaskingTestMixin,
)


@tests.skip_data
@tests.iristest_timing_decorator
class TestBroadcasting(
    tests.IrisTest_nometa, CubeArithmeticBroadcastingTestMixin
):
    @property
    def data_op(self):
        return operator.add

    @property
    def cube_func(self):
        return add


@tests.iristest_timing_decorator
class TestMasking(tests.IrisTest_nometa, CubeArithmeticMaskingTestMixin):
    @property
    def data_op(self):
        return operator.add

    @property
    def cube_func(self):
        return add


class TestCoordMatch(CubeArithmeticCoordsTest):
    def test_no_match(self):
        cube1, cube2 = self.SetUpNonMatching()
        with self.assertRaises(ValueError):
            add(cube1, cube2)

    def test_reversed_points(self):
        cube1, cube2 = self.SetUpReversed()
        with self.assertRaises(ValueError):
            add(cube1, cube2)


@tests.iristest_timing_decorator
class TestMaskedConstant(
    tests.IrisTest_nometa, CubeArithmeticMaskedConstantTestMixin
):
    @property
    def data_op(self):
        return operator.add

    @property
    def cube_func(self):
        return add


if __name__ == "__main__":
    tests.main()



"""
Global Average Annual Temperature Maps
======================================

Produces maps of global temperature forecasts from the A1B and E1 scenarios.

The data used comes from the HadGEM2-AO model simulations for the A1B and E1
scenarios, both of which were derived using the IMAGE Integrated Assessment
Model (Johns et al. 2011; Lowe et al. 2009).

References
----------

    Johns T.C., et al. (2011) Climate change under aggressive mitigation: the
    ENSEMBLES multi-model experiment. Climate Dynamics, Vol 37, No. 9-10,
    doi:10.1007/s00382-011-1005-5.

    Lowe J.A., C.D. Hewitt, D.P. Van Vuuren, T.C. Johns, E. Stehfest, J-F.
    Royer, and P. van der Linden, 2009. New Study For Climate Modeling,
    Analyses, and Scenarios. Eos Trans. AGU, Vol 90, No. 21,
    doi:10.1029/2009EO210001.

"""

import os.path

import matplotlib.pyplot as plt
import numpy as np

import iris
import iris.coords as coords
import iris.plot as iplt


def cop_metadata_callback(cube, field, filename):
    """
    A function which adds an "Experiment" coordinate which comes from the
    filename.
    """

    # Extract the experiment name (such as a1b or e1) from the filename (in
    # this case it is just the parent folder's name)
    containing_folder = os.path.dirname(filename)
    experiment_label = os.path.basename(containing_folder)

    # Create a coordinate with the experiment label in it
    exp_coord = coords.AuxCoord(
        experiment_label, long_name="Experiment", units="no_unit"
    )

    # and add it to the cube
    cube.add_aux_coord(exp_coord)


def main():
    # Load e1 and a1 using the callback to update the metadata
    e1 = iris.load_cube(
        iris.sample_data_path("E1.2098.pp"), callback=cop_metadata_callback
    )
    a1b = iris.load_cube(
        iris.sample_data_path("A1B.2098.pp"), callback=cop_metadata_callback
    )

    # Load the global average data and add an 'Experiment' coord it
    global_avg = iris.load_cube(iris.sample_data_path("pre-industrial.pp"))

    # Define evenly spaced contour levels: -2.5, -1.5, ... 15.5, 16.5 with the
    # specific colours
    levels = np.arange(20) - 2.5
    red = (
        np.array(
            [
                0,
                0,
                221,
                239,
                229,
                217,
                239,
                234,
                228,
                222,
                205,
                196,
                161,
                137,
                116,
                89,
                77,
                60,
                51,
            ]
        )
        / 256.0
    )
    green = (
        np.array(
            [
                16,
                217,
                242,
                243,
                235,
                225,
                190,
                160,
                128,
                87,
                72,
                59,
                33,
                21,
                29,
                30,
                30,
                29,
                26,
            ]
        )
        / 256.0
    )
    blue = (
        np.array(
            [
                255,
                255,
                243,
                169,
                99,
                51,
                63,
                37,
                39,
                21,
                27,
                23,
                22,
                26,
                29,
                28,
                27,
                25,
                22,
            ]
        )
        / 256.0
    )

    # Put those colours into an array which can be passed to contourf as the
    # specific colours for each level
    colors = np.array([red, green, blue]).T

    # Subtract the global

    # Iterate over each latitude longitude slice for both e1 and a1b scenarios
    # simultaneously
    for e1_slice, a1b_slice in zip(
        e1.slices(["latitude", "longitude"]),
        a1b.slices(["latitude", "longitude"]),
    ):

        time_coord = a1b_slice.coord("time")

        # Calculate the difference from the mean
        delta_e1 = e1_slice - global_avg
        delta_a1b = a1b_slice - global_avg

        # Make a wider than normal figure to house two maps side-by-side
        fig = plt.figure(figsize=(12, 5))

        # Get the time datetime from the coordinate
        time = time_coord.units.num2date(time_coord.points[0])
        # Set a title for the entire figure, giving the time in a nice format
        # of "MonthName Year". Also, set the y value for the title so that it
        # is not tight to the top of the plot.
        fig.suptitle(
            "Annual Temperature Predictions for " + time.strftime("%Y"),
            y=0.9,
            fontsize=18,
        )

        # Add the first subplot showing the E1 scenario
        plt.subplot(121)
        plt.title("HadGEM2 E1 Scenario", fontsize=10)
        iplt.contourf(delta_e1, levels, colors=colors, extend="both")
        plt.gca().coastlines()
        # get the current axes' subplot for use later on
        plt1_ax = plt.gca()

        # Add the second subplot showing the A1B scenario
        plt.subplot(122)
        plt.title("HadGEM2 A1B-Image Scenario", fontsize=10)
        contour_result = iplt.contourf(
            delta_a1b, levels, colors=colors, extend="both"
        )
        plt.gca().coastlines()
        # get the current axes' subplot for use later on
        plt2_ax = plt.gca()

        # Now add a colourbar who's leftmost point is the same as the leftmost
        # point of the left hand plot and rightmost point is the rightmost
        # point of the right hand plot

        # Get the positions of the 2nd plot and the left position of the 1st
        # plot
        left, bottom, width, height = plt2_ax.get_position().bounds
        first_plot_left = plt1_ax.get_position().bounds[0]

        # the width of the colorbar should now be simple
        width = left - first_plot_left + width

        # Add axes to the figure, to place the colour bar
        colorbar_axes = fig.add_axes([first_plot_left, 0.18, width, 0.03])

        # Add the colour bar
        cbar = plt.colorbar(
            contour_result, colorbar_axes, orientation="horizontal"
        )

        # Label the colour bar and add ticks
        cbar.set_label(e1_slice.units)
        cbar.ax.tick_params(length=0)

        iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.analysis._regrid.CurvilinearRegridder`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.analysis.cartography import rotate_pole
from iris.cube import Cube
from iris.coords import AuxCoord, DimCoord
from iris.coord_systems import GeogCS, RotatedGeogCS
from iris.fileformats.pp import EARTH_RADIUS
from iris.analysis._regrid import CurvilinearRegridder as Regridder
from iris.tests.stock import global_pp, lat_lon_cube


RESULT_DIR = ("analysis", "regrid")


class Test___init__(tests.IrisTest):
    def setUp(self):
        self.src_grid = lat_lon_cube()
        self.bad = np.ones((3, 4))
        self.weights = np.ones(self.src_grid.shape, self.src_grid.dtype)

    def test_bad_src_type(self):
        with self.assertRaisesRegex(TypeError, "'src_grid_cube'"):
            Regridder(self.bad, self.src_grid, self.weights)

    def test_bad_grid_type(self):
        with self.assertRaisesRegex(TypeError, "'target_grid_cube'"):
            Regridder(self.src_grid, self.bad, self.weights)


@tests.skip_data
class Test___call__(tests.IrisTest):
    def setUp(self):
        self.func_setup = (
            "iris.analysis._regrid."
            "_regrid_weighted_curvilinear_to_rectilinear__prepare"
        )
        self.func_operate = (
            "iris.analysis._regrid."
            "_regrid_weighted_curvilinear_to_rectilinear__perform"
        )
        # Define a test source grid and target grid, basically the same.
        self.src_grid = global_pp()
        self.tgt_grid = global_pp()
        # Modify the names so we can tell them apart.
        self.src_grid.rename("src_grid")
        self.tgt_grid.rename("TARGET_GRID")
        # Replace the source-grid x and y coords with equivalent 2d versions.
        x_coord = self.src_grid.coord("longitude")
        y_coord = self.src_grid.coord("latitude")
        (nx,) = x_coord.shape
        (ny,) = y_coord.shape
        xx, yy = np.meshgrid(x_coord.points, y_coord.points)
        self.src_grid.remove_coord(x_coord)
        self.src_grid.remove_coord(y_coord)
        x_coord_2d = AuxCoord(
            xx,
            standard_name=x_coord.standard_name,
            units=x_coord.units,
            coord_system=x_coord.coord_system,
        )
        y_coord_2d = AuxCoord(
            yy,
            standard_name=y_coord.standard_name,
            units=y_coord.units,
            coord_system=y_coord.coord_system,
        )
        self.src_grid.add_aux_coord(x_coord_2d, (0, 1))
        self.src_grid.add_aux_coord(y_coord_2d, (0, 1))
        self.weights = np.ones(self.src_grid.shape, self.src_grid.dtype)
        # Define an actual, dummy cube for the internal partial result, so we
        # can do a cubelist merge on it, which is too complicated to mock out.
        self.dummy_slice_result = Cube([1])

    def test_same_src_as_init(self):
        # Check the regridder call calls the underlying routines as expected.
        src_grid = self.src_grid
        target_grid = self.tgt_grid
        regridder = Regridder(src_grid, target_grid, self.weights)
        with mock.patch(
            self.func_setup, return_value=mock.sentinel.regrid_info
        ) as patch_setup:
            with mock.patch(
                self.func_operate, return_value=self.dummy_slice_result
            ) as patch_operate:
                result = regridder(src_grid)
        patch_setup.assert_called_once_with(
            src_grid, self.weights, target_grid
        )
        patch_operate.assert_called_once_with(
            src_grid, mock.sentinel.regrid_info
        )
        # The result is a re-merged version of the internal result, so it is
        # therefore '==' but not the same object.
        self.assertEqual(result, self.dummy_slice_result)

    def test_no_weights(self):
        # Check we can use the regridder without weights.
        src_grid = self.src_grid
        target_grid = self.tgt_grid
        regridder = Regridder(src_grid, target_grid)
        with mock.patch(
            self.func_setup, return_value=mock.sentinel.regrid_info
        ) as patch_setup:
            with mock.patch(
                self.func_operate, return_value=self.dummy_slice_result
            ):
                _ = regridder(src_grid)
        patch_setup.assert_called_once_with(src_grid, None, target_grid)

    def test_diff_src_from_init(self):
        # Check we can call the regridder with a different cube from the one we
        # built it with.
        src_grid = self.src_grid
        target_grid = self.tgt_grid
        regridder = Regridder(src_grid, target_grid, self.weights)
        # Provide a "different" cube for the actual regrid.
        different_src_cube = self.src_grid.copy()
        # Rename so we can distinguish them.
        different_src_cube.rename("Different_source")
        with mock.patch(
            self.func_setup, return_value=mock.sentinel.regrid_info
        ):
            with mock.patch(
                self.func_operate, return_value=self.dummy_slice_result
            ) as patch_operate:
                _ = regridder(different_src_cube)
        patch_operate.assert_called_once_with(
            different_src_cube, mock.sentinel.regrid_info
        )

    def test_caching(self):
        # Check that it calculates regrid info just once, and re-uses it in
        # subsequent calls.
        src_grid = self.src_grid
        target_grid = self.tgt_grid
        regridder = Regridder(src_grid, target_grid, self.weights)
        different_src_cube = self.src_grid.copy()
        different_src_cube.rename("Different_source")
        with mock.patch(
            self.func_setup, return_value=mock.sentinel.regrid_info
        ) as patch_setup:
            with mock.patch(
                self.func_operate, return_value=self.dummy_slice_result
            ) as patch_operate:
                _ = regridder(src_grid)
                _ = regridder(different_src_cube)
        patch_setup.assert_called_once_with(
            src_grid, self.weights, target_grid
        )
        self.assertEqual(len(patch_operate.call_args_list), 2)
        self.assertEqual(
            patch_operate.call_args_list,
            [
                mock.call(src_grid, mock.sentinel.regrid_info),
                mock.call(different_src_cube, mock.sentinel.regrid_info),
            ],
        )


@tests.skip_data
class Test___call____bad_src(tests.IrisTest):
    def setUp(self):
        self.src_grid = global_pp()
        y = self.src_grid.coord("latitude")
        x = self.src_grid.coord("longitude")
        self.src_grid.remove_coord("latitude")
        self.src_grid.remove_coord("longitude")
        self.src_grid.add_aux_coord(y, 0)
        self.src_grid.add_aux_coord(x, 1)
        weights = np.ones(self.src_grid.shape, self.src_grid.dtype)
        self.regridder = Regridder(self.src_grid, self.src_grid, weights)

    def test_bad_src_type(self):
        with self.assertRaisesRegex(TypeError, "must be a Cube"):
            self.regridder(np.ones((3, 4)))

    def test_bad_src_shape(self):
        with self.assertRaisesRegex(
            ValueError, "not defined on the same source grid"
        ):
            self.regridder(self.src_grid[::2, ::2])


class Test__call__multidimensional(tests.IrisTest):
    def test_multidim(self):
        # Testing with >2D data to demonstrate correct operation over
        # additional non-XY dimensions (including data masking), which is
        # handled by the PointInCell wrapper class.

        # Define a simple target grid first, in plain latlon coordinates.
        plain_latlon_cs = GeogCS(EARTH_RADIUS)
        grid_x_coord = DimCoord(
            points=[15.0, 25.0, 35.0],
            bounds=[[10.0, 20.0], [20.0, 30.0], [30.0, 40.0]],
            standard_name="longitude",
            units="degrees",
            coord_system=plain_latlon_cs,
        )
        grid_y_coord = DimCoord(
            points=[-30.0, -50.0],
            bounds=[[-20.0, -40.0], [-40.0, -60.0]],
            standard_name="latitude",
            units="degrees",
            coord_system=plain_latlon_cs,
        )
        grid_cube = Cube(np.zeros((2, 3)))
        grid_cube.add_dim_coord(grid_y_coord, 0)
        grid_cube.add_dim_coord(grid_x_coord, 1)

        # Define some key points in true-lat/lon thta have known positions
        # First 3x2 points in the centre of each output cell.
        x_centres, y_centres = np.meshgrid(
            grid_x_coord.points, grid_y_coord.points
        )
        # An extra point also falling in cell 1, 1
        x_in11, y_in11 = 26.3, -48.2
        # An extra point completely outside the target grid
        x_out, y_out = 70.0, -40.0

        # Define a rotated coord system for the source data
        pole_lon, pole_lat = -125.3, 53.4
        src_cs = RotatedGeogCS(
            grid_north_pole_latitude=pole_lat,
            grid_north_pole_longitude=pole_lon,
            ellipsoid=plain_latlon_cs,
        )

        # Concatenate all the testpoints in a flat array, and find the rotated
        # equivalents.
        xx = list(x_centres.flat[:]) + [x_in11, x_out]
        yy = list(y_centres.flat[:]) + [y_in11, y_out]
        xx, yy = rotate_pole(
            lons=np.array(xx),
            lats=np.array(yy),
            pole_lon=pole_lon,
            pole_lat=pole_lat,
        )
        # Define handy index numbers for all these.
        i00, i01, i02, i10, i11, i12, i_in, i_out = range(8)

        # Build test data in the shape Z,YX = (3, 8)
        data = [
            [1, 2, 3, 11, 12, 13, 7, 99],
            [1, 2, 3, 11, 12, 13, 7, 99],
            [7, 6, 5, 51, 52, 53, 12, 1],
        ]
        mask = [
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
        ]
        src_data = np.ma.array(data, mask=mask, dtype=float)

        # Make the source cube.
        src_cube = Cube(src_data)
        src_x = AuxCoord(
            xx,
            standard_name="grid_longitude",
            units="degrees",
            coord_system=src_cs,
        )
        src_y = AuxCoord(
            yy,
            standard_name="grid_latitude",
            units="degrees",
            coord_system=src_cs,
        )
        src_z = DimCoord(np.arange(3), long_name="z")
        src_cube.add_dim_coord(src_z, 0)
        src_cube.add_aux_coord(src_x, 1)
        src_cube.add_aux_coord(src_y, 1)
        # Add in some extra metadata, to ensure it gets copied over.
        src_cube.add_aux_coord(DimCoord([0], long_name="extra_scalar_coord"))
        src_cube.attributes["extra_attr"] = 12.3

        # Define what the expected answers should be, shaped (3, 2, 3).
        expected_result = [
            [[1.0, 2.0, 3.0], [11.0, 0.5 * (12 + 7), 13.0]],
            [[1.0, -999, 3.0], [11.0, 12.0, 13.0]],
            [[7.0, 6.0, 5.0], [51.0, 0.5 * (52 + 12), 53.0]],
        ]
        expected_result = np.ma.masked_less(expected_result, 0)

        # Perform the calculation with the regridder.
        regridder = Regridder(src_cube, grid_cube)

        # Check all is as expected.
        result = regridder(src_cube)
        self.assertEqual(result.coord("z"), src_cube.coord("z"))
        self.assertEqual(
            result.coord("extra_scalar_coord"),
            src_cube.coord("extra_scalar_coord"),
        )
        self.assertEqual(
            result.coord("longitude"), grid_cube.coord("longitude")
        )
        self.assertEqual(result.coord("latitude"), grid_cube.coord("latitude"))
        self.assertMaskedArrayAlmostEqual(result.data, expected_result)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the intersection of Coords

"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import numpy as np

import iris
import iris.cube
import iris.coord_systems
import iris.coords
import iris.tests.stock


class TestCubeIntersectTheoretical(tests.IrisTest):
    def test_simple_intersect(self):
        cube = iris.cube.Cube(
            np.array(
                [
                    [1, 2, 3, 4, 5],
                    [2, 3, 4, 5, 6],
                    [3, 4, 5, 6, 7],
                    [4, 5, 6, 7, 8],
                    [5, 6, 7, 8, 9],
                ],
                dtype=np.int32,
            )
        )

        lonlat_cs = iris.coord_systems.RotatedGeogCS(10, 20)
        cube.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(5, dtype=np.float32) * 90 - 180,
                "longitude",
                units="degrees",
                coord_system=lonlat_cs,
            ),
            1,
        )
        cube.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(5, dtype=np.float32) * 45 - 90,
                "latitude",
                units="degrees",
                coord_system=lonlat_cs,
            ),
            0,
        )
        cube.add_aux_coord(
            iris.coords.DimCoord(
                points=np.int32(11), long_name="pressure", units="Pa"
            )
        )
        cube.rename("temperature")
        cube.units = "K"

        cube2 = iris.cube.Cube(
            np.array(
                [
                    [1, 2, 3, 4, 5],
                    [2, 3, 4, 5, 6],
                    [3, 4, 5, 6, 7],
                    [4, 5, 6, 7, 8],
                    [5, 6, 7, 8, 50],
                ],
                dtype=np.int32,
            )
        )

        lonlat_cs = iris.coord_systems.RotatedGeogCS(10, 20)
        cube2.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(5, dtype=np.float32) * 90,
                "longitude",
                units="degrees",
                coord_system=lonlat_cs,
            ),
            1,
        )
        cube2.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(5, dtype=np.float32) * 45 - 90,
                "latitude",
                units="degrees",
                coord_system=lonlat_cs,
            ),
            0,
        )
        cube2.add_aux_coord(
            iris.coords.DimCoord(
                points=np.int32(11), long_name="pressure", units="Pa"
            )
        )
        cube2.rename("")

        r = iris.analysis.maths.intersection_of_cubes(cube, cube2)
        self.assertCML(r, ("cdm", "test_simple_cube_intersection.cml"))


class TestCoordIntersect(tests.IrisTest):
    def test_commutative(self):
        step = 4.0
        c1 = iris.coords.DimCoord(np.arange(100) * step)
        offset_points = c1.points.copy()
        offset_points -= step * 30
        c2 = c1.copy(points=offset_points)

        i1 = c1.intersect(c2)
        i2 = c2.intersect(c1)
        self.assertEqual(i1, i2)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.fileformat.ff.FFHeader`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import collections
from unittest import mock

import numpy as np

from iris.fileformats._ff import FFHeader


MyGrid = collections.namedtuple("MyGrid", "column row real horiz_grid_type")


class Test_grid(tests.IrisTest):
    def _header(self, grid_staggering):
        with mock.patch.object(
            FFHeader, "__init__", mock.Mock(return_value=None)
        ):
            header = FFHeader()
        header.grid_staggering = grid_staggering
        header.column_dependent_constants = mock.sentinel.column
        header.row_dependent_constants = mock.sentinel.row
        header.real_constants = mock.sentinel.real
        header.horiz_grid_type = mock.sentinel.horiz_grid_type
        return header

    def _test_grid_staggering(self, grid_staggering):
        header = self._header(grid_staggering)
        with mock.patch.dict(
            FFHeader.GRID_STAGGERING_CLASS, {grid_staggering: MyGrid}
        ):
            grid = header.grid()
        self.assertIsInstance(grid, MyGrid)
        self.assertIs(grid.column, mock.sentinel.column)
        self.assertIs(grid.row, mock.sentinel.row)
        self.assertIs(grid.real, mock.sentinel.real)
        self.assertIs(grid.horiz_grid_type, mock.sentinel.horiz_grid_type)

    def test_new_dynamics(self):
        self._test_grid_staggering(3)

    def test_end_game(self):
        self._test_grid_staggering(6)

    def test_unknown(self):
        header = self._header(0)
        with mock.patch(
            "iris.fileformats._ff.NewDynamics",
            mock.Mock(return_value=mock.sentinel.grid),
        ):
            with mock.patch("warnings.warn") as warn:
                grid = header.grid()
        warn.assert_called_with(
            "Staggered grid type: 0 not currently"
            " interpreted, assuming standard C-grid"
        )
        self.assertIs(grid, mock.sentinel.grid)


@tests.skip_data
class Test_integer_constants(tests.IrisTest):
    def test_read_ints(self):
        test_file_path = tests.get_data_path(("FF", "structured", "small"))
        ff_header = FFHeader(test_file_path)
        self.assertEqual(ff_header.integer_constants.dtype, np.dtype(">i8"))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

from sphinx.ext.autodoc import DataDocumenter, ModuleLevelDocumenter
try:
    # Use 'object_description' in place of the former 'safe_repr' function.
    from sphinx.util.inspect import object_description as safe_repr
except ImportError:
    # 'safe_repr' is the old usage, for Sphinx<1.3.
    from sphinx.util.inspect import safe_repr

from iris.analysis import Aggregator


class IrisDataDocumenter(DataDocumenter):
    priority = 100

    def add_directive_header(self, sig):
        ModuleLevelDocumenter.add_directive_header(self, sig)
        if not self.options.annotation:
            try:
                objrepr = safe_repr(self.object)
            except ValueError:
                pass
            else:
                self.add_line(u'   :annotation:', '<autodoc>')
        elif self.options.annotation is object():
            pass
        else:
            self.add_line(
                u'   :annotation: {}'.format(self.options.annotation),
                '<autodoc>')


def handler(app, what, name, obj, options, signature, return_annotation):
    if what == 'data':
        if isinstance(obj, object) and issubclass(obj.__class__, Aggregator):
            signature = '()'
            return_annotation = '{} instance.'.format(obj.__class__.__name__)
    return signature, return_annotation


def setup(app):
    app.add_autodocumenter(IrisDataDocumenter, override=True)
    app.connect('autodoc-process-signature', handler)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests the high-level plotting interface.

"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests
import iris.tests.test_plot as test_plot

import iris

# Run tests in no graphics mode if matplotlib is not available.
if tests.MPL_AVAILABLE:
    import matplotlib.pyplot as plt
    import iris.plot as iplt
    import iris.quickplot as qplt


# Caches _load_theta so subsequent calls are faster
def cache(fn, cache={}):
    def inner(*args, **kwargs):
        key = "result"
        if not cache:
            cache[key] = fn(*args, **kwargs)
        return cache[key]

    return inner


@cache
def _load_theta():
    path = tests.get_data_path(("PP", "COLPEX", "theta_and_orog_subset.pp"))
    theta = iris.load_cube(path, "air_potential_temperature")

    # Improve the unit
    theta.units = "K"

    return theta


@tests.skip_data
@tests.skip_plot
class TestQuickplotCoordinatesGiven(test_plot.TestPlotCoordinatesGiven):
    def setUp(self):
        tests.GraphicsTest.setUp(self)
        filename = tests.get_data_path(
            ("PP", "COLPEX", "theta_and_orog_subset.pp")
        )
        self.cube = test_plot.load_cube_once(
            filename, "air_potential_temperature"
        )

        self.draw_module = iris.quickplot
        self.contourf = test_plot.LambdaStr(
            "iris.quickplot.contourf",
            lambda cube, *args, **kwargs: iris.quickplot.contourf(
                cube, *args, **kwargs
            ),
        )
        self.contour = test_plot.LambdaStr(
            "iris.quickplot.contour",
            lambda cube, *args, **kwargs: iris.quickplot.contour(
                cube, *args, **kwargs
            ),
        )
        self.points = test_plot.LambdaStr(
            "iris.quickplot.points",
            lambda cube, *args, **kwargs: iris.quickplot.points(
                cube, c=cube.data, *args, **kwargs
            ),
        )
        self.plot = test_plot.LambdaStr(
            "iris.quickplot.plot",
            lambda cube, *args, **kwargs: iris.quickplot.plot(
                cube, *args, **kwargs
            ),
        )

        self.results = {
            "yx": (
                [self.contourf, ["grid_latitude", "grid_longitude"]],
                [self.contourf, ["grid_longitude", "grid_latitude"]],
                [self.contour, ["grid_latitude", "grid_longitude"]],
                [self.contour, ["grid_longitude", "grid_latitude"]],
                [self.points, ["grid_latitude", "grid_longitude"]],
                [self.points, ["grid_longitude", "grid_latitude"]],
            ),
            "zx": (
                [self.contourf, ["model_level_number", "grid_longitude"]],
                [self.contourf, ["grid_longitude", "model_level_number"]],
                [self.contour, ["model_level_number", "grid_longitude"]],
                [self.contour, ["grid_longitude", "model_level_number"]],
                [self.points, ["model_level_number", "grid_longitude"]],
                [self.points, ["grid_longitude", "model_level_number"]],
            ),
            "tx": (
                [self.contourf, ["time", "grid_longitude"]],
                [self.contourf, ["grid_longitude", "time"]],
                [self.contour, ["time", "grid_longitude"]],
                [self.contour, ["grid_longitude", "time"]],
                [self.points, ["time", "grid_longitude"]],
                [self.points, ["grid_longitude", "time"]],
            ),
            "x": ([self.plot, ["grid_longitude"]],),
            "y": ([self.plot, ["grid_latitude"]],),
        }


@tests.skip_data
@tests.skip_plot
class TestLabels(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.theta = _load_theta()

    def _slice(self, coords):
        """Returns the first cube containing the requested coordinates."""
        for cube in self.theta.slices(coords):
            break
        return cube

    def _small(self):
        # Use a restricted size so we can make out the detail
        cube = self._slice(["model_level_number", "grid_longitude"])
        return cube[:5, :5]

    def test_contour(self):
        qplt.contour(self._small())
        self.check_graphic()

        qplt.contourf(
            self._small(), coords=["model_level_number", "grid_longitude"]
        )
        self.check_graphic()

    def test_contourf(self):
        qplt.contourf(self._small())

        cube = self._small()
        iplt.orography_at_points(cube)

        self.check_graphic()

        qplt.contourf(
            self._small(), coords=["model_level_number", "grid_longitude"]
        )
        self.check_graphic()

        qplt.contourf(
            self._small(), coords=["grid_longitude", "model_level_number"]
        )
        self.check_graphic()

    def test_contourf_axes_specified(self):
        # Check that the contourf function does not modify the matplotlib
        # pyplot state machine.

        # Create a figure and axes to be used by contourf
        plt.figure()
        axes1 = plt.axes()

        # Create test figure and axes which will be the new results
        # of plt.gcf and plt.gca.
        plt.figure()
        axes2 = plt.axes()

        # Add a title to the test axes.
        plt.title("This should not be changed")
        # Draw the contourf on a specific axes.
        qplt.contourf(self._small(), axes=axes1)

        # Ensure that the correct axes got the appropriate title.
        self.assertEqual(axes2.get_title(), "This should not be changed")
        self.assertEqual(axes1.get_title(), "Air potential temperature")

        # Check that the axes labels were set correctly.
        self.assertEqual(axes1.get_xlabel(), "Grid longitude / degrees")
        self.assertEqual(axes1.get_ylabel(), "Altitude / m")

    def test_contourf_nameless(self):
        cube = self._small()
        cube.standard_name = None
        cube.attributes["STASH"] = ""
        qplt.contourf(cube, coords=["grid_longitude", "model_level_number"])
        self.check_graphic()

    def test_pcolor(self):
        qplt.pcolor(self._small())
        self.check_graphic()

    def test_pcolormesh(self):
        qplt.pcolormesh(self._small())

        # cube = self._small()
        # iplt.orography_at_bounds(cube)

        self.check_graphic()

    def test_pcolormesh_str_symbol(self):
        pcube = self._small().copy()
        pcube.coords("level_height")[0].units = "centimeters"
        qplt.pcolormesh(pcube)

        self.check_graphic()

    def test_map(self):
        cube = self._slice(["grid_latitude", "grid_longitude"])
        qplt.contour(cube)
        self.check_graphic()

        # check that the result of adding 360 to the data is *almost* identically the same result
        lon = cube.coord("grid_longitude")
        lon.points = lon.points + 360
        qplt.contour(cube)
        self.check_graphic()

    def test_alignment(self):
        cube = self._small()
        qplt.contourf(cube)
        # qplt.outline(cube)
        qplt.points(cube)
        self.check_graphic()


@tests.skip_data
@tests.skip_plot
class TestTimeReferenceUnitsLabels(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        path = tests.get_data_path(("PP", "aPProt1", "rotatedMHtimecube.pp"))
        self.cube = iris.load_cube(path)[:, 0, 0]

    def test_reference_time_units(self):
        # units should not be displayed for a reference time
        qplt.plot(self.cube.coord("time"), self.cube)
        plt.gcf().autofmt_xdate()
        self.check_graphic()

    def test_not_reference_time_units(self):
        # units should be displayed for other time coordinates
        qplt.plot(self.cube.coord("forecast_period"), self.cube)
        self.check_graphic()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.build_cube_metadata`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    get_attr_units


class TestGetAttrUnits(tests.IrisTest):
    @staticmethod
    def _make_cf_var(global_attributes=None):
        if global_attributes is None:
            global_attributes = {}

        cf_group = mock.Mock(global_attributes=global_attributes)

        cf_var = mock.MagicMock(
            cf_name='sound_frequency',
            cf_data=mock.Mock(spec=[]),
            standard_name=None,
            long_name=None,
            units=u'\u266b',
            dtype=np.float64,
            cell_methods=None,
            cf_group=cf_group)
        return cf_var

    def test_unicode_character(self):
        attributes = {}
        expected_attributes = {'invalid_units': u'\u266b'}
        cf_var = self._make_cf_var()
        attr_units = get_attr_units(cf_var, attributes)
        self.assertEqual(attr_units, '?')
        self.assertEqual(attributes, expected_attributes)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.fileformat.ff.ENDGame`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.fileformats._ff import ENDGame


class Test(tests.IrisTest):
    def test_class_attributes(self):
        reals = np.arange(6) + 100
        grid = ENDGame(None, None, reals, None)
        self.assertEqual(grid._v_offset, -0.5)


class Test__y_vectors(tests.IrisTest):
    def _test(self, row, yp, yv):
        reals = np.arange(6) + 100
        grid = ENDGame(None, row, reals, None)
        result_yp, result_yv = grid._y_vectors()
        self.assertArrayEqual(result_yp, yp)
        self.assertArrayEqual(result_yv, yv)

    def test_none(self):
        self._test(row=None, yp=None, yv=None)

    def test_1d(self):
        self._test(
            row=np.array([[0], [1], [2], [3]]), yp=np.array([0, 1, 2]), yv=None
        )

    def test_2d(self):
        self._test(
            row=np.array([[0, 0], [1, 10], [2, 20], [3, 30]]),
            yp=np.array([0, 1, 2]),
            yv=np.array([0, 10, 20, 30]),
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.broadcast_to_shape`."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris.util import broadcast_to_shape


class Test_broadcast_to_shape(tests.IrisTest):
    def test_same_shape(self):
        # broadcast to current shape should result in no change
        a = np.random.random([2, 3])
        b = broadcast_to_shape(a, a.shape, (0, 1))
        self.assertArrayEqual(b, a)

    def test_added_dimensions(self):
        # adding two dimensions, on at the front and one in the middle of
        # the existing dimensions
        a = np.random.random([2, 3])
        b = broadcast_to_shape(a, (5, 2, 4, 3), (1, 3))
        for i in range(5):
            for j in range(4):
                self.assertArrayEqual(b[i, :, j, :], a)

    def test_added_dimensions_transpose(self):
        # adding dimensions and having the dimensions of the input
        # transposed
        a = np.random.random([2, 3])
        b = broadcast_to_shape(a, (5, 3, 4, 2), (3, 1))
        for i in range(5):
            for j in range(4):
                self.assertArrayEqual(b[i, :, j, :].T, a)

    def test_masked(self):
        # masked arrays are also accepted
        a = np.random.random([2, 3])
        m = ma.array(a, mask=[[0, 1, 0], [0, 1, 1]])
        b = broadcast_to_shape(m, (5, 3, 4, 2), (3, 1))
        for i in range(5):
            for j in range(4):
                self.assertMaskedArrayEqual(b[i, :, j, :].T, m)

    def test_masked_degenerate(self):
        # masked arrays can have degenerate masks too
        a = np.random.random([2, 3])
        m = ma.array(a)
        b = broadcast_to_shape(m, (5, 3, 4, 2), (3, 1))
        for i in range(5):
            for j in range(4):
                self.assertMaskedArrayEqual(b[i, :, j, :].T, m)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :func:`iris.analysis.maths.divide` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import operator

from iris.analysis.maths import divide
from iris.cube import Cube
from iris.tests.unit.analysis.maths import (
    CubeArithmeticBroadcastingTestMixin,
    CubeArithmeticMaskingTestMixin,
    CubeArithmeticCoordsTest,
)


@tests.skip_data
@tests.iristest_timing_decorator
class TestBroadcasting(
    tests.IrisTest_nometa, CubeArithmeticBroadcastingTestMixin
):
    @property
    def data_op(self):
        return operator.truediv

    @property
    def cube_func(self):
        return divide


@tests.iristest_timing_decorator
class TestMasking(tests.IrisTest_nometa, CubeArithmeticMaskingTestMixin):
    @property
    def data_op(self):
        return operator.truediv

    @property
    def cube_func(self):
        return divide

    def test_unmasked_div_zero(self):
        # Ensure cube behaviour matches numpy operator behaviour for the
        # handling of arrays containing 0.
        dat_a = np.array([0.0, 0.0, 0.0, 0.0])
        dat_b = np.array([2.0, 2.0, 2.0, 2.0])

        cube_a = Cube(dat_a)
        cube_b = Cube(dat_b)

        com = self.data_op(dat_b, dat_a)
        res = self.cube_func(cube_b, cube_a).data

        self.assertArrayEqual(com, res)

    def test_masked_div_zero(self):
        # Ensure cube behaviour matches numpy operator behaviour for the
        # handling of arrays containing 0.
        dat_a = np.ma.array([0.0, 0.0, 0.0, 0.0], mask=False)
        dat_b = np.ma.array([2.0, 2.0, 2.0, 2.0], mask=False)

        cube_a = Cube(dat_a)
        cube_b = Cube(dat_b)

        com = self.data_op(dat_b, dat_a)
        res = self.cube_func(cube_b, cube_a).data

        self.assertMaskedArrayEqual(com, res, strict=True)


class TestCoordMatch(CubeArithmeticCoordsTest):
    def test_no_match(self):
        cube1, cube2 = self.SetUpNonMatching()
        with self.assertRaises(ValueError):
            divide(cube1, cube2)

    def test_reversed_points(self):
        cube1, cube2 = self.SetUpReversed()
        with self.assertRaises(ValueError):
            divide(cube1, cube2)


if __name__ == "__main__":
    tests.main()



"""
Test Data Showing Inset Plots
=============================

This example demonstrates the use of a single 3D data cube with time, latitude
and longitude dimensions to plot a temperature series for a single latitude
coordinate, with an inset plot of the data region.

"""

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np

import iris
import iris.quickplot as qplt
import iris.plot as iplt


def main():
    cube1 = iris.load_cube(iris.sample_data_path("ostia_monthly.nc"))
    # Slice into cube to retrieve data for the inset map showing the
    # data region
    region = cube1[-1, :, :]
    # Average over latitude to reduce cube to 1 dimension
    plot_line = region.collapsed("latitude", iris.analysis.MEAN)

    # Open a window for plotting
    fig = plt.figure()
    # Add a single subplot (axes). Could also use "ax_main = plt.subplot()"
    ax_main = fig.add_subplot(1, 1, 1)
    # Produce a quick plot of the 1D cube
    qplt.plot(plot_line)

    # Set x limits to match the data
    ax_main.set_xlim(0, plot_line.coord("longitude").points.max())
    # Adjust the y limits so that the inset map won't clash with main plot
    ax_main.set_ylim(294, 310)
    ax_main.set_title("Meridional Mean Temperature")
    # Add grid lines
    ax_main.grid()

    # Add a second set of axes specifying the fractional coordinates within
    # the figure with bottom left corner at x=0.55, y=0.58 with width
    # 0.3 and height 0.25.
    # Also specify the projection
    ax_sub = fig.add_axes(
        [0.55, 0.58, 0.3, 0.25],
        projection=ccrs.Mollweide(central_longitude=180),
    )

    # Use iris.plot (iplt) here so colour bar properties can be specified
    # Also use a sequential colour scheme to reduce confusion for those with
    # colour-blindness
    iplt.pcolormesh(region, cmap="Blues")
    # Manually set the orientation and tick marks on your colour bar
    ticklist = np.linspace(np.min(region.data), np.max(region.data), 4)
    plt.colorbar(orientation="horizontal", ticks=ticklist)
    ax_sub.set_title("Data Region")
    # Add coastlines
    ax_sub.coastlines()
    # request to show entire map, using the colour mesh on the data region only
    ax_sub.set_global()

    qplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris.fileformats.netcdf.parse_cell_methods`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

from iris.coords import CellMethod
from iris.fileformats.netcdf import parse_cell_methods


class Test(tests.IrisTest):
    def test_simple(self):
        cell_method_strings = [
            "time: mean",
            "time : mean",
        ]
        expected = (CellMethod(method="mean", coords="time"),)
        for cell_method_str in cell_method_strings:
            res = parse_cell_methods(cell_method_str)
            self.assertEqual(res, expected)

    def test_with_interval(self):
        cell_method_strings = [
            "time: variance (interval: 1 hr)",
            "time : variance (interval: 1 hr)",
        ]
        expected = (
            CellMethod(method="variance", coords="time", intervals="1 hr"),
        )
        for cell_method_str in cell_method_strings:
            res = parse_cell_methods(cell_method_str)
            self.assertEqual(res, expected)

    def test_multiple(self):
        cell_method_strings = [
            "time: maximum (interval: 1 hr) time: mean (interval: 1 day)",
            "time : maximum (interval: 1 hr) time: mean (interval: 1 day)",
            "time: maximum (interval: 1 hr) time : mean (interval: 1 day)",
            "time : maximum (interval: 1 hr) time : mean (interval: 1 day)",
        ]
        expected = (
            CellMethod(method="maximum", coords="time", intervals="1 hr"),
            CellMethod(method="mean", coords="time", intervals="1 day"),
        )
        for cell_method_str in cell_method_strings:
            res = parse_cell_methods(cell_method_str)
            self.assertEqual(res, expected)

    def test_comment(self):
        cell_method_strings = [
            "time: maximum (interval: 1 hr comment: first bit) "
            "time: mean (interval: 1 day comment: second bit)",
            "time : maximum (interval: 1 hr comment: first bit) "
            "time: mean (interval: 1 day comment: second bit)",
            "time: maximum (interval: 1 hr comment: first bit) "
            "time : mean (interval: 1 day comment: second bit)",
            "time : maximum (interval: 1 hr comment: first bit) "
            "time : mean (interval: 1 day comment: second bit)",
        ]
        expected = (
            CellMethod(
                method="maximum",
                coords="time",
                intervals="1 hr",
                comments="first bit",
            ),
            CellMethod(
                method="mean",
                coords="time",
                intervals="1 day",
                comments="second bit",
            ),
        )
        for cell_method_str in cell_method_strings:
            res = parse_cell_methods(cell_method_str)
            self.assertEqual(res, expected)

    def test_portions_of_cells(self):
        cell_method_strings = [
            "area: mean where sea_ice over sea",
            "area : mean where sea_ice over sea",
        ]
        expected = (
            CellMethod(method="mean where sea_ice over sea", coords="area"),
        )
        for cell_method_str in cell_method_strings:
            res = parse_cell_methods(cell_method_str)
            self.assertEqual(res, expected)

    def test_climatology(self):
        cell_method_strings = [
            "time: minimum within days time: mean over days",
            "time : minimum within days time: mean over days",
            "time: minimum within days time : mean over days",
            "time : minimum within days time : mean over days",
        ]
        expected = (
            CellMethod(method="minimum within days", coords="time"),
            CellMethod(method="mean over days", coords="time"),
        )
        for cell_method_str in cell_method_strings:
            res = parse_cell_methods(cell_method_str)
            self.assertEqual(res, expected)

    def test_climatology_with_unknown_method(self):
        cell_method_strings = [
            "time: min within days time: mean over days",
            "time : min within days time: mean over days",
            "time: min within days time : mean over days",
            "time : min within days time : mean over days",
        ]
        expected = (
            CellMethod(method="min within days", coords="time"),
            CellMethod(method="mean over days", coords="time"),
        )
        for cell_method_str in cell_method_strings:
            with mock.patch("warnings.warn") as warn:
                res = parse_cell_methods(cell_method_str)
            self.assertIn(
                "NetCDF variable contains unknown cell method 'min'",
                warn.call_args[0][0],
            )
            self.assertEqual(res, expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Integration tests for concatenating cubes with differing time coord epochs
using :func:`iris.util.unify_time_units`.

"""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import cf_units
import numpy as np

import iris.coords
from iris._concatenate import concatenate
import iris.cube
import iris.tests.stock as stock
from iris.util import unify_time_units


class Test_concatenate__epoch(tests.IrisTest):
    def simple_1d_time_cubes(self, reftimes, coords_points):
        cubes = []
        data_points = [273, 275, 278, 277, 274]
        for reftime, coord_points in zip(reftimes, coords_points):
            cube = iris.cube.Cube(
                np.array(data_points, dtype=np.float32),
                standard_name="air_temperature",
                units="K",
            )
            unit = cf_units.Unit(reftime, calendar="gregorian")
            coord = iris.coords.DimCoord(
                points=np.array(coord_points, dtype=np.float32),
                standard_name="time",
                units=unit,
            )
            cube.add_dim_coord(coord, 0)
            cubes.append(cube)
        return cubes

    def test_concat_1d_with_differing_time_units(self):
        reftimes = [
            "hours since 1970-01-01 00:00:00",
            "hours since 1970-01-02 00:00:00",
        ]
        coords_points = [[1, 2, 3, 4, 5], [1, 2, 3, 4, 5]]
        cubes = self.simple_1d_time_cubes(reftimes, coords_points)
        unify_time_units(cubes)
        result = concatenate(cubes)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (10,))


class Test_cubes_with_aux_coord(tests.IrisTest):
    def create_cube(self):
        data = np.arange(4).reshape(2, 2)

        lat = iris.coords.DimCoord(
            [0, 30], standard_name="latitude", units="degrees"
        )
        lon = iris.coords.DimCoord(
            [0, 15], standard_name="longitude", units="degrees"
        )
        height = iris.coords.AuxCoord([1.5], standard_name="height", units="m")
        t_unit = cf_units.Unit(
            "hours since 1970-01-01 00:00:00", calendar="gregorian"
        )
        time = iris.coords.DimCoord([0, 6], standard_name="time", units=t_unit)

        cube = iris.cube.Cube(data, standard_name="air_temperature", units="K")
        cube.add_dim_coord(time, 0)
        cube.add_dim_coord(lat, 1)
        cube.add_aux_coord(lon, 1)
        cube.add_aux_coord(height)
        return cube

    def test_diff_aux_coord(self):
        cube_a = self.create_cube()
        cube_b = cube_a.copy()
        cube_b.coord("time").points = [12, 18]
        cube_b.coord("longitude").points = [120, 150]

        result = concatenate([cube_a, cube_b])
        self.assertEqual(len(result), 2)

    def test_ignore_diff_aux_coord(self):
        cube_a = self.create_cube()
        cube_b = cube_a.copy()
        cube_b.coord("time").points = [12, 18]
        cube_b.coord("longitude").points = [120, 150]

        result = concatenate([cube_a, cube_b], check_aux_coords=False)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 2))


class Test_cubes_with_cell_measure(tests.IrisTest):
    def create_cube(self):
        data = np.arange(4).reshape(2, 2)

        lat = iris.coords.DimCoord(
            [0, 30], standard_name="latitude", units="degrees"
        )
        volume = iris.coords.CellMeasure(
            [0, 15], measure="volume", long_name="volume"
        )
        area = iris.coords.CellMeasure(
            [1.5], standard_name="height", units="m"
        )
        t_unit = cf_units.Unit(
            "hours since 1970-01-01 00:00:00", calendar="gregorian"
        )
        time = iris.coords.DimCoord([0, 6], standard_name="time", units=t_unit)

        cube = iris.cube.Cube(data, standard_name="air_temperature", units="K")
        cube.add_dim_coord(time, 0)
        cube.add_dim_coord(lat, 1)
        cube.add_cell_measure(volume, 1)
        cube.add_cell_measure(area)
        return cube

    def test_diff_cell_measure(self):
        cube_a = self.create_cube()
        cube_b = cube_a.copy()
        cube_b.coord("time").points = [12, 18]
        cube_b.cell_measure("volume").data = [120, 150]

        result = concatenate([cube_a, cube_b])
        self.assertEqual(len(result), 2)

    def test_ignore_diff_cell_measure(self):
        cube_a = self.create_cube()
        cube_b = cube_a.copy()
        cube_b.coord("time").points = [12, 18]
        cube_b.cell_measure("volume").data = [120, 150]

        result = concatenate([cube_a, cube_b], check_cell_measures=False)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 2))


class Test_cubes_with_ancillary_variables(tests.IrisTest):
    def create_cube(self):
        data = np.arange(4).reshape(2, 2)

        lat = iris.coords.DimCoord(
            [0, 30], standard_name="latitude", units="degrees"
        )
        quality = iris.coords.AncillaryVariable([0, 15], long_name="quality")
        height = iris.coords.AncillaryVariable(
            [1.5], standard_name="height", units="m"
        )
        t_unit = cf_units.Unit(
            "hours since 1970-01-01 00:00:00", calendar="gregorian"
        )
        time = iris.coords.DimCoord([0, 6], standard_name="time", units=t_unit)

        cube = iris.cube.Cube(data, standard_name="air_temperature", units="K")
        cube.add_dim_coord(time, 0)
        cube.add_dim_coord(lat, 1)
        cube.add_ancillary_variable(quality, 1)
        cube.add_ancillary_variable(height)
        return cube

    def test_diff_ancillary_variables(self):
        cube_a = self.create_cube()
        cube_b = cube_a.copy()
        cube_b.coord("time").points = [12, 18]
        cube_b.ancillary_variable("quality").data = [120, 150]

        result = concatenate([cube_a, cube_b])
        self.assertEqual(len(result), 2)

    def test_ignore_diff_ancillary_variables(self):
        cube_a = self.create_cube()
        cube_b = cube_a.copy()
        cube_b.coord("time").points = [12, 18]
        cube_b.ancillary_variable("quality").data = [120, 150]

        result = concatenate([cube_a, cube_b], check_ancils=False)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].shape, (4, 2))


class Test_anonymous_dims(tests.IrisTest):
    def setUp(self):
        data = np.arange(12).reshape(2, 3, 2)
        self.cube = iris.cube.Cube(
            data, standard_name="air_temperature", units="K"
        )

        # Time coord
        t_unit = cf_units.Unit(
            "hours since 1970-01-01 00:00:00", calendar="gregorian"
        )
        t_coord = iris.coords.DimCoord(
            [0, 6], standard_name="time", units=t_unit
        )
        self.cube.add_dim_coord(t_coord, 0)

        # Lats and lons
        self.x_coord = iris.coords.DimCoord(
            [15, 30], standard_name="longitude", units="degrees"
        )
        self.y_coord = iris.coords.DimCoord(
            [0, 30, 60], standard_name="latitude", units="degrees"
        )
        self.x_coord_2D = iris.coords.AuxCoord(
            [[0, 15], [30, 45], [60, 75]],
            standard_name="longitude",
            units="degrees",
        )
        self.y_coord_non_monotonic = iris.coords.AuxCoord(
            [0, 30, 15], standard_name="latitude", units="degrees"
        )

    def test_matching_2d_longitudes(self):
        cube1 = self.cube
        cube1.add_dim_coord(self.y_coord, 1)
        cube1.add_aux_coord(self.x_coord_2D, (1, 2))

        cube2 = cube1.copy()
        cube2.coord("time").points = [12, 18]
        result = concatenate([cube1, cube2])
        self.assertEqual(len(result), 1)

    def test_differing_2d_longitudes(self):
        cube1 = self.cube
        cube1.add_aux_coord(self.y_coord, 1)
        cube1.add_aux_coord(self.x_coord_2D, (1, 2))

        cube2 = cube1.copy()
        cube2.coord("time").points = [12, 18]
        cube2.coord("longitude").points = [[-30, -15], [0, 15], [30, 45]]

        result = concatenate([cube1, cube2])
        self.assertEqual(len(result), 2)

    def test_matching_non_monotonic_latitudes(self):
        cube1 = self.cube
        cube1.add_aux_coord(self.y_coord_non_monotonic, 1)
        cube1.add_aux_coord(self.x_coord, 2)

        cube2 = cube1.copy()
        cube2.coord("time").points = [12, 18]

        result = concatenate([cube1, cube2])
        self.assertEqual(len(result), 1)

    def test_differing_non_monotonic_latitudes(self):
        cube1 = self.cube
        cube1.add_aux_coord(self.y_coord_non_monotonic, 1)
        cube1.add_aux_coord(self.x_coord, 2)

        cube2 = cube1.copy()
        cube2.coord("time").points = [12, 18]
        cube2.coord("latitude").points = [30, 0, 15]

        result = concatenate([cube1, cube2])
        self.assertEqual(len(result), 2)

    def test_concatenate_along_anon_dim(self):
        cube1 = self.cube
        cube1.add_aux_coord(self.y_coord_non_monotonic, 1)
        cube1.add_aux_coord(self.x_coord, 2)

        cube2 = cube1.copy()
        cube2.coord("latitude").points = [30, 0, 15]

        result = concatenate([cube1, cube2])
        self.assertEqual(len(result), 2)


class Test_anonymous_dims_alternate_mapping(tests.IrisTest):
    # Ensure that anonymous concatenation is not sensitive to dimension mapping
    # of the anonymous dimension.
    def setUp(self):
        self.cube = stock.simple_3d()
        coord = self.cube.coord("wibble")
        self.cube.remove_coord(coord)
        self.cube.add_aux_coord(coord, 0)

    def test_concatenate_anom_1st_dim(self):
        # Check that concatenation along a non anonymous dimension is
        # insensitive to the dimension which is anonymous.
        # Concatenate along longitude.
        # DIM: cube(--, lat, lon)   & cube(--, lat, lon')
        # AUX: cube(wibble, --, --) & cube(wibble, --, --)
        cube1 = self.cube[..., :2]
        cube2 = self.cube[..., 2:]
        result = concatenate([cube1, cube2])
        self.assertEqual(len(result), 1)

    def test_concatenate_anom_2nd_dim(self):
        # Check that concatenation along a non anonymous dimension is
        # insensitive to the dimension which is anonymous.
        # Concatenate along longitude.
        # DIM: cube(lon, --, lat)   & cube(lon', ---, lat)
        # AUX: cube(--, wibble, --) & cube(--, wibble, --)
        cube1 = self.cube[..., :2]
        cube2 = self.cube[..., 2:]
        cube1.transpose((2, 0, 1))
        cube2.transpose((2, 0, 1))
        result = concatenate([cube1, cube2])
        self.assertEqual(len(result), 1)

    def test_concatenate_anom_3rd_dim(self):
        # Check that concatenation along a non anonymous dimension is
        # insensitive to the dimension which is anonymous.
        # Concatenate along longitude.
        # DIM: cube(lat, lon, --)   & cube(lat, lon', --)
        # AUX: cube(--, --, wibble) & cube(--, --, wibble)
        cube1 = self.cube[..., :2]
        cube2 = self.cube[..., 2:]
        cube1.transpose((1, 2, 0))
        cube2.transpose((1, 2, 0))
        result = concatenate([cube1, cube2])
        self.assertEqual(len(result), 1)


if __name__ == "__main__":
    tests.main()
